# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import proto  # type: ignore

from arrivy.google.cloud.bigquery_v2.types import encryption_config
from arrivy.google.cloud.bigquery_v2.types import model_reference as gcb_model_reference
from arrivy.google.cloud.bigquery_v2.types import standard_sql
from arrivy.google.cloud.bigquery_v2.types import table_reference
from google.protobuf import timestamp_pb2  # type: ignore
from google.protobuf import wrappers_pb2  # type: ignore


__protobuf__ = proto.module(
    package="arrivy.google.cloud.bigquery.v2",
    manifest={
        "Model",
        "GetModelRequest",
        "PatchModelRequest",
        "DeleteModelRequest",
        "ListModelsRequest",
        "ListModelsResponse",
    },
)


class Model(proto.Message):
    r"""

    Attributes:
        etag (str):
            Output only. A hash of this resource.
        model_reference (arrivy.google.cloud.bigquery_v2.types.ModelReference):
            Required. Unique identifier for this model.
        creation_time (int):
            Output only. The time when this model was
            created, in millisecs since the epoch.
        last_modified_time (int):
            Output only. The time when this model was
            last modified, in millisecs since the epoch.
        description (str):
            Optional. A user-friendly description of this
            model.
        friendly_name (str):
            Optional. A descriptive name for this model.
        labels (Mapping[str, str]):
            The labels associated with this model. You
            can use these to organize and group your models.
            Label keys and values can be no longer than 63
            characters, can only contain lowercase letters,
            numeric characters, underscores and dashes.
            International characters are allowed. Label
            values are optional. Label keys must start with
            a letter and each label in the list must have a
            different key.
        expiration_time (int):
            Optional. The time when this model expires,
            in milliseconds since the epoch. If not present,
            the model will persist indefinitely. Expired
            models will be deleted and their storage
            reclaimed.  The defaultTableExpirationMs
            property of the encapsulating dataset can be
            used to set a default expirationTime on newly
            created models.
        location (str):
            Output only. The geographic location where
            the model resides. This value is inherited from
            the dataset.
        encryption_configuration (arrivy.google.cloud.bigquery_v2.types.EncryptionConfiguration):
            Custom encryption configuration (e.g., Cloud
            KMS keys). This shows the encryption
            configuration of the model data while stored in
            BigQuery storage. This field can be used with
            PatchModel to update encryption key for an
            already encrypted model.
        model_type (arrivy.google.cloud.bigquery_v2.types.Model.ModelType):
            Output only. Type of the model resource.
        training_runs (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun]):
            Output only. Information for all training runs in increasing
            order of start_time.
        feature_columns (Sequence[arrivy.google.cloud.bigquery_v2.types.StandardSqlField]):
            Output only. Input feature columns that were
            used to train this model.
        label_columns (Sequence[arrivy.google.cloud.bigquery_v2.types.StandardSqlField]):
            Output only. Label columns that were used to train this
            model. The output of the model will have a `predicted_`
            prefix to these columns.
        best_trial_id (int):
            The best trial_id across all training runs.
    """

    class ModelType(proto.Enum):
        r"""Indicates the type of the Model."""
        MODEL_TYPE_UNSPECIFIED = 0
        LINEAR_REGRESSION = 1
        LOGISTIC_REGRESSION = 2
        KMEANS = 3
        MATRIX_FACTORIZATION = 4
        DNN_CLASSIFIER = 5
        TENSORFLOW = 6
        DNN_REGRESSOR = 7
        BOOSTED_TREE_REGRESSOR = 9
        BOOSTED_TREE_CLASSIFIER = 10
        ARIMA = 11
        AUTOML_REGRESSOR = 12
        AUTOML_CLASSIFIER = 13
        ARIMA_PLUS = 19

    class LossType(proto.Enum):
        r"""Loss metric to evaluate model training performance."""
        LOSS_TYPE_UNSPECIFIED = 0
        MEAN_SQUARED_LOSS = 1
        MEAN_LOG_LOSS = 2

    class DistanceType(proto.Enum):
        r"""Distance metric used to compute the distance between two
        points.
        """
        DISTANCE_TYPE_UNSPECIFIED = 0
        EUCLIDEAN = 1
        COSINE = 2

    class DataSplitMethod(proto.Enum):
        r"""Indicates the method to split input data into multiple
        tables.
        """
        DATA_SPLIT_METHOD_UNSPECIFIED = 0
        RANDOM = 1
        CUSTOM = 2
        SEQUENTIAL = 3
        NO_SPLIT = 4
        AUTO_SPLIT = 5

    class DataFrequency(proto.Enum):
        r"""Type of supported data frequency for time series forecasting
        models.
        """
        DATA_FREQUENCY_UNSPECIFIED = 0
        AUTO_FREQUENCY = 1
        YEARLY = 2
        QUARTERLY = 3
        MONTHLY = 4
        WEEKLY = 5
        DAILY = 6
        HOURLY = 7
        PER_MINUTE = 8

    class HolidayRegion(proto.Enum):
        r"""Type of supported holiday regions for time series forecasting
        models.
        """
        HOLIDAY_REGION_UNSPECIFIED = 0
        GLOBAL = 1
        NA = 2
        JAPAC = 3
        EMEA = 4
        LAC = 5
        AE = 6
        AR = 7
        AT = 8
        AU = 9
        BE = 10
        BR = 11
        CA = 12
        CH = 13
        CL = 14
        CN = 15
        CO = 16
        CS = 17
        CZ = 18
        DE = 19
        DK = 20
        DZ = 21
        EC = 22
        EE = 23
        EG = 24
        ES = 25
        FI = 26
        FR = 27
        GB = 28
        GR = 29
        HK = 30
        HU = 31
        ID = 32
        IE = 33
        IL = 34
        IN = 35
        IR = 36
        IT = 37
        JP = 38
        KR = 39
        LV = 40
        MA = 41
        MX = 42
        MY = 43
        NG = 44
        NL = 45
        NO = 46
        NZ = 47
        PE = 48
        PH = 49
        PK = 50
        PL = 51
        PT = 52
        RO = 53
        RS = 54
        RU = 55
        SA = 56
        SE = 57
        SG = 58
        SI = 59
        SK = 60
        TH = 61
        TR = 62
        TW = 63
        UA = 64
        US = 65
        VE = 66
        VN = 67
        ZA = 68

    class LearnRateStrategy(proto.Enum):
        r"""Indicates the learning rate optimization strategy to use."""
        LEARN_RATE_STRATEGY_UNSPECIFIED = 0
        LINE_SEARCH = 1
        CONSTANT = 2

    class OptimizationStrategy(proto.Enum):
        r"""Indicates the optimization strategy used for training."""
        OPTIMIZATION_STRATEGY_UNSPECIFIED = 0
        BATCH_GRADIENT_DESCENT = 1
        NORMAL_EQUATION = 2

    class FeedbackType(proto.Enum):
        r"""Indicates the training algorithm to use for matrix
        factorization models.
        """
        FEEDBACK_TYPE_UNSPECIFIED = 0
        IMPLICIT = 1
        EXPLICIT = 2

    class SeasonalPeriod(proto.Message):
        r""" """

        class SeasonalPeriodType(proto.Enum):
            r""""""
            SEASONAL_PERIOD_TYPE_UNSPECIFIED = 0
            NO_SEASONALITY = 1
            DAILY = 2
            WEEKLY = 3
            MONTHLY = 4
            QUARTERLY = 5
            YEARLY = 6

    class KmeansEnums(proto.Message):
        r""" """

        class KmeansInitializationMethod(proto.Enum):
            r"""Indicates the method used to initialize the centroids for
            KMeans clustering algorithm.
            """
            KMEANS_INITIALIZATION_METHOD_UNSPECIFIED = 0
            RANDOM = 1
            CUSTOM = 2
            KMEANS_PLUS_PLUS = 3

    class RegressionMetrics(proto.Message):
        r"""Evaluation metrics for regression and explicit feedback type
        matrix factorization models.

        Attributes:
            mean_absolute_error (google.protobuf.wrappers_pb2.DoubleValue):
                Mean absolute error.
            mean_squared_error (google.protobuf.wrappers_pb2.DoubleValue):
                Mean squared error.
            mean_squared_log_error (google.protobuf.wrappers_pb2.DoubleValue):
                Mean squared log error.
            median_absolute_error (google.protobuf.wrappers_pb2.DoubleValue):
                Median absolute error.
            r_squared (google.protobuf.wrappers_pb2.DoubleValue):
                R^2 score. This corresponds to r2_score in ML.EVALUATE.
        """

        mean_absolute_error = proto.Field(
            proto.MESSAGE,
            number=1,
            message=wrappers_pb2.DoubleValue,
        )
        mean_squared_error = proto.Field(
            proto.MESSAGE,
            number=2,
            message=wrappers_pb2.DoubleValue,
        )
        mean_squared_log_error = proto.Field(
            proto.MESSAGE,
            number=3,
            message=wrappers_pb2.DoubleValue,
        )
        median_absolute_error = proto.Field(
            proto.MESSAGE,
            number=4,
            message=wrappers_pb2.DoubleValue,
        )
        r_squared = proto.Field(
            proto.MESSAGE,
            number=5,
            message=wrappers_pb2.DoubleValue,
        )

    class AggregateClassificationMetrics(proto.Message):
        r"""Aggregate metrics for classification/classifier models. For
        multi-class models, the metrics are either macro-averaged or
        micro-averaged. When macro-averaged, the metrics are calculated
        for each label and then an unweighted average is taken of those
        values. When micro-averaged, the metric is calculated globally
        by counting the total number of correctly predicted rows.

        Attributes:
            precision (google.protobuf.wrappers_pb2.DoubleValue):
                Precision is the fraction of actual positive
                predictions that had positive actual labels. For
                multiclass this is a macro-averaged metric
                treating each class as a binary classifier.
            recall (google.protobuf.wrappers_pb2.DoubleValue):
                Recall is the fraction of actual positive
                labels that were given a positive prediction.
                For multiclass this is a macro-averaged metric.
            accuracy (google.protobuf.wrappers_pb2.DoubleValue):
                Accuracy is the fraction of predictions given
                the correct label. For multiclass this is a
                micro-averaged metric.
            threshold (google.protobuf.wrappers_pb2.DoubleValue):
                Threshold at which the metrics are computed.
                For binary classification models this is the
                positive class threshold. For multi-class
                classfication models this is the confidence
                threshold.
            f1_score (google.protobuf.wrappers_pb2.DoubleValue):
                The F1 score is an average of recall and
                precision. For multiclass this is a
                macro-averaged metric.
            log_loss (google.protobuf.wrappers_pb2.DoubleValue):
                Logarithmic Loss. For multiclass this is a
                macro-averaged metric.
            roc_auc (google.protobuf.wrappers_pb2.DoubleValue):
                Area Under a ROC Curve. For multiclass this
                is a macro-averaged metric.
        """

        precision = proto.Field(
            proto.MESSAGE,
            number=1,
            message=wrappers_pb2.DoubleValue,
        )
        recall = proto.Field(
            proto.MESSAGE,
            number=2,
            message=wrappers_pb2.DoubleValue,
        )
        accuracy = proto.Field(
            proto.MESSAGE,
            number=3,
            message=wrappers_pb2.DoubleValue,
        )
        threshold = proto.Field(
            proto.MESSAGE,
            number=4,
            message=wrappers_pb2.DoubleValue,
        )
        f1_score = proto.Field(
            proto.MESSAGE,
            number=5,
            message=wrappers_pb2.DoubleValue,
        )
        log_loss = proto.Field(
            proto.MESSAGE,
            number=6,
            message=wrappers_pb2.DoubleValue,
        )
        roc_auc = proto.Field(
            proto.MESSAGE,
            number=7,
            message=wrappers_pb2.DoubleValue,
        )

    class BinaryClassificationMetrics(proto.Message):
        r"""Evaluation metrics for binary classification/classifier
        models.

        Attributes:
            aggregate_classification_metrics (arrivy.google.cloud.bigquery_v2.types.Model.AggregateClassificationMetrics):
                Aggregate classification metrics.
            binary_confusion_matrix_list (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.BinaryClassificationMetrics.BinaryConfusionMatrix]):
                Binary confusion matrix at multiple
                thresholds.
            positive_label (str):
                Label representing the positive class.
            negative_label (str):
                Label representing the negative class.
        """

        class BinaryConfusionMatrix(proto.Message):
            r"""Confusion matrix for binary classification models.

            Attributes:
                positive_class_threshold (google.protobuf.wrappers_pb2.DoubleValue):
                    Threshold value used when computing each of
                    the following metric.
                true_positives (google.protobuf.wrappers_pb2.Int64Value):
                    Number of true samples predicted as true.
                false_positives (google.protobuf.wrappers_pb2.Int64Value):
                    Number of false samples predicted as true.
                true_negatives (google.protobuf.wrappers_pb2.Int64Value):
                    Number of true samples predicted as false.
                false_negatives (google.protobuf.wrappers_pb2.Int64Value):
                    Number of false samples predicted as false.
                precision (google.protobuf.wrappers_pb2.DoubleValue):
                    The fraction of actual positive predictions
                    that had positive actual labels.
                recall (google.protobuf.wrappers_pb2.DoubleValue):
                    The fraction of actual positive labels that
                    were given a positive prediction.
                f1_score (google.protobuf.wrappers_pb2.DoubleValue):
                    The equally weighted average of recall and
                    precision.
                accuracy (google.protobuf.wrappers_pb2.DoubleValue):
                    The fraction of predictions given the correct
                    label.
            """

            positive_class_threshold = proto.Field(
                proto.MESSAGE,
                number=1,
                message=wrappers_pb2.DoubleValue,
            )
            true_positives = proto.Field(
                proto.MESSAGE,
                number=2,
                message=wrappers_pb2.Int64Value,
            )
            false_positives = proto.Field(
                proto.MESSAGE,
                number=3,
                message=wrappers_pb2.Int64Value,
            )
            true_negatives = proto.Field(
                proto.MESSAGE,
                number=4,
                message=wrappers_pb2.Int64Value,
            )
            false_negatives = proto.Field(
                proto.MESSAGE,
                number=5,
                message=wrappers_pb2.Int64Value,
            )
            precision = proto.Field(
                proto.MESSAGE,
                number=6,
                message=wrappers_pb2.DoubleValue,
            )
            recall = proto.Field(
                proto.MESSAGE,
                number=7,
                message=wrappers_pb2.DoubleValue,
            )
            f1_score = proto.Field(
                proto.MESSAGE,
                number=8,
                message=wrappers_pb2.DoubleValue,
            )
            accuracy = proto.Field(
                proto.MESSAGE,
                number=9,
                message=wrappers_pb2.DoubleValue,
            )

        aggregate_classification_metrics = proto.Field(
            proto.MESSAGE,
            number=1,
            message="Model.AggregateClassificationMetrics",
        )
        binary_confusion_matrix_list = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message="Model.BinaryClassificationMetrics.BinaryConfusionMatrix",
        )
        positive_label = proto.Field(
            proto.STRING,
            number=3,
        )
        negative_label = proto.Field(
            proto.STRING,
            number=4,
        )

    class MultiClassClassificationMetrics(proto.Message):
        r"""Evaluation metrics for multi-class classification/classifier
        models.

        Attributes:
            aggregate_classification_metrics (arrivy.google.cloud.bigquery_v2.types.Model.AggregateClassificationMetrics):
                Aggregate classification metrics.
            confusion_matrix_list (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.MultiClassClassificationMetrics.ConfusionMatrix]):
                Confusion matrix at different thresholds.
        """

        class ConfusionMatrix(proto.Message):
            r"""Confusion matrix for multi-class classification models.

            Attributes:
                confidence_threshold (google.protobuf.wrappers_pb2.DoubleValue):
                    Confidence threshold used when computing the
                    entries of the confusion matrix.
                rows (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.MultiClassClassificationMetrics.ConfusionMatrix.Row]):
                    One row per actual label.
            """

            class Entry(proto.Message):
                r"""A single entry in the confusion matrix.

                Attributes:
                    predicted_label (str):
                        The predicted label. For confidence_threshold > 0, we will
                        also add an entry indicating the number of items under the
                        confidence threshold.
                    item_count (google.protobuf.wrappers_pb2.Int64Value):
                        Number of items being predicted as this
                        label.
                """

                predicted_label = proto.Field(
                    proto.STRING,
                    number=1,
                )
                item_count = proto.Field(
                    proto.MESSAGE,
                    number=2,
                    message=wrappers_pb2.Int64Value,
                )

            class Row(proto.Message):
                r"""A single row in the confusion matrix.

                Attributes:
                    actual_label (str):
                        The original label of this row.
                    entries (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.MultiClassClassificationMetrics.ConfusionMatrix.Entry]):
                        Info describing predicted label distribution.
                """

                actual_label = proto.Field(
                    proto.STRING,
                    number=1,
                )
                entries = proto.RepeatedField(
                    proto.MESSAGE,
                    number=2,
                    message="Model.MultiClassClassificationMetrics.ConfusionMatrix.Entry",
                )

            confidence_threshold = proto.Field(
                proto.MESSAGE,
                number=1,
                message=wrappers_pb2.DoubleValue,
            )
            rows = proto.RepeatedField(
                proto.MESSAGE,
                number=2,
                message="Model.MultiClassClassificationMetrics.ConfusionMatrix.Row",
            )

        aggregate_classification_metrics = proto.Field(
            proto.MESSAGE,
            number=1,
            message="Model.AggregateClassificationMetrics",
        )
        confusion_matrix_list = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message="Model.MultiClassClassificationMetrics.ConfusionMatrix",
        )

    class ClusteringMetrics(proto.Message):
        r"""Evaluation metrics for clustering models.

        Attributes:
            davies_bouldin_index (google.protobuf.wrappers_pb2.DoubleValue):
                Davies-Bouldin index.
            mean_squared_distance (google.protobuf.wrappers_pb2.DoubleValue):
                Mean of squared distances between each sample
                to its cluster centroid.
            clusters (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ClusteringMetrics.Cluster]):
                Information for all clusters.
        """

        class Cluster(proto.Message):
            r"""Message containing the information about one cluster.

            Attributes:
                centroid_id (int):
                    Centroid id.
                feature_values (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ClusteringMetrics.Cluster.FeatureValue]):
                    Values of highly variant features for this
                    cluster.
                count (google.protobuf.wrappers_pb2.Int64Value):
                    Count of training data rows that were
                    assigned to this cluster.
            """

            class FeatureValue(proto.Message):
                r"""Representative value of a single feature within the cluster.

                This message has `oneof`_ fields (mutually exclusive fields).
                For each oneof, at most one member field can be set at the same time.
                Setting any member of the oneof automatically clears all other
                members.

                .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

                Attributes:
                    feature_column (str):
                        The feature column name.
                    numerical_value (google.protobuf.wrappers_pb2.DoubleValue):
                        The numerical feature value. This is the
                        centroid value for this feature.

                        This field is a member of `oneof`_ ``value``.
                    categorical_value (arrivy.google.cloud.bigquery_v2.types.Model.ClusteringMetrics.Cluster.FeatureValue.CategoricalValue):
                        The categorical feature value.

                        This field is a member of `oneof`_ ``value``.
                """

                class CategoricalValue(proto.Message):
                    r"""Representative value of a categorical feature.

                    Attributes:
                        category_counts (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ClusteringMetrics.Cluster.FeatureValue.CategoricalValue.CategoryCount]):
                            Counts of all categories for the categorical feature. If
                            there are more than ten categories, we return top ten (by
                            count) and return one more CategoryCount with category
                            "*OTHER*" and count as aggregate counts of remaining
                            categories.
                    """

                    class CategoryCount(proto.Message):
                        r"""Represents the count of a single category within the cluster.

                        Attributes:
                            category (str):
                                The name of category.
                            count (google.protobuf.wrappers_pb2.Int64Value):
                                The count of training samples matching the
                                category within the cluster.
                        """

                        category = proto.Field(
                            proto.STRING,
                            number=1,
                        )
                        count = proto.Field(
                            proto.MESSAGE,
                            number=2,
                            message=wrappers_pb2.Int64Value,
                        )

                    category_counts = proto.RepeatedField(
                        proto.MESSAGE,
                        number=1,
                        message="Model.ClusteringMetrics.Cluster.FeatureValue.CategoricalValue.CategoryCount",
                    )

                feature_column = proto.Field(
                    proto.STRING,
                    number=1,
                )
                numerical_value = proto.Field(
                    proto.MESSAGE,
                    number=2,
                    oneof="value",
                    message=wrappers_pb2.DoubleValue,
                )
                categorical_value = proto.Field(
                    proto.MESSAGE,
                    number=3,
                    oneof="value",
                    message="Model.ClusteringMetrics.Cluster.FeatureValue.CategoricalValue",
                )

            centroid_id = proto.Field(
                proto.INT64,
                number=1,
            )
            feature_values = proto.RepeatedField(
                proto.MESSAGE,
                number=2,
                message="Model.ClusteringMetrics.Cluster.FeatureValue",
            )
            count = proto.Field(
                proto.MESSAGE,
                number=3,
                message=wrappers_pb2.Int64Value,
            )

        davies_bouldin_index = proto.Field(
            proto.MESSAGE,
            number=1,
            message=wrappers_pb2.DoubleValue,
        )
        mean_squared_distance = proto.Field(
            proto.MESSAGE,
            number=2,
            message=wrappers_pb2.DoubleValue,
        )
        clusters = proto.RepeatedField(
            proto.MESSAGE,
            number=3,
            message="Model.ClusteringMetrics.Cluster",
        )

    class RankingMetrics(proto.Message):
        r"""Evaluation metrics used by weighted-ALS models specified by
        feedback_type=implicit.

        Attributes:
            mean_average_precision (google.protobuf.wrappers_pb2.DoubleValue):
                Calculates a precision per user for all the
                items by ranking them and then averages all the
                precisions across all the users.
            mean_squared_error (google.protobuf.wrappers_pb2.DoubleValue):
                Similar to the mean squared error computed in
                regression and explicit recommendation models
                except instead of computing the rating directly,
                the output from evaluate is computed against a
                preference which is 1 or 0 depending on if the
                rating exists or not.
            normalized_discounted_cumulative_gain (google.protobuf.wrappers_pb2.DoubleValue):
                A metric to determine the goodness of a
                ranking calculated from the predicted confidence
                by comparing it to an ideal rank measured by the
                original ratings.
            average_rank (google.protobuf.wrappers_pb2.DoubleValue):
                Determines the goodness of a ranking by
                computing the percentile rank from the predicted
                confidence and dividing it by the original rank.
        """

        mean_average_precision = proto.Field(
            proto.MESSAGE,
            number=1,
            message=wrappers_pb2.DoubleValue,
        )
        mean_squared_error = proto.Field(
            proto.MESSAGE,
            number=2,
            message=wrappers_pb2.DoubleValue,
        )
        normalized_discounted_cumulative_gain = proto.Field(
            proto.MESSAGE,
            number=3,
            message=wrappers_pb2.DoubleValue,
        )
        average_rank = proto.Field(
            proto.MESSAGE,
            number=4,
            message=wrappers_pb2.DoubleValue,
        )

    class ArimaForecastingMetrics(proto.Message):
        r"""Model evaluation metrics for ARIMA forecasting models.

        Attributes:
            non_seasonal_order (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ArimaOrder]):
                Non-seasonal order.
            arima_fitting_metrics (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ArimaFittingMetrics]):
                Arima model fitting metrics.
            seasonal_periods (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.SeasonalPeriod.SeasonalPeriodType]):
                Seasonal periods. Repeated because multiple
                periods are supported for one time series.
            has_drift (Sequence[bool]):
                Whether Arima model fitted with drift or not.
                It is always false when d is not 1.
            time_series_id (Sequence[str]):
                Id to differentiate different time series for
                the large-scale case.
            arima_single_model_forecasting_metrics (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.ArimaForecastingMetrics.ArimaSingleModelForecastingMetrics]):
                Repeated as there can be many metric sets
                (one for each model) in auto-arima and the
                large-scale case.
        """

        class ArimaSingleModelForecastingMetrics(proto.Message):
            r"""Model evaluation metrics for a single ARIMA forecasting
            model.

            Attributes:
                non_seasonal_order (arrivy.google.cloud.bigquery_v2.types.Model.ArimaOrder):
                    Non-seasonal order.
                arima_fitting_metrics (arrivy.google.cloud.bigquery_v2.types.Model.ArimaFittingMetrics):
                    Arima fitting metrics.
                has_drift (bool):
                    Is arima model fitted with drift or not. It
                    is always false when d is not 1.
                time_series_id (str):
                    The time_series_id value for this time series. It will be
                    one of the unique values from the time_series_id_column
                    specified during ARIMA model training. Only present when
                    time_series_id_column training option was used.
                time_series_ids (Sequence[str]):
                    The tuple of time_series_ids identifying this time series.
                    It will be one of the unique tuples of values present in the
                    time_series_id_columns specified during ARIMA model
                    training. Only present when time_series_id_columns training
                    option was used and the order of values here are same as the
                    order of time_series_id_columns.
                seasonal_periods (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.SeasonalPeriod.SeasonalPeriodType]):
                    Seasonal periods. Repeated because multiple
                    periods are supported for one time series.
                has_holiday_effect (google.protobuf.wrappers_pb2.BoolValue):
                    If true, holiday_effect is a part of time series
                    decomposition result.
                has_spikes_and_dips (google.protobuf.wrappers_pb2.BoolValue):
                    If true, spikes_and_dips is a part of time series
                    decomposition result.
                has_step_changes (google.protobuf.wrappers_pb2.BoolValue):
                    If true, step_changes is a part of time series decomposition
                    result.
            """

            non_seasonal_order = proto.Field(
                proto.MESSAGE,
                number=1,
                message="Model.ArimaOrder",
            )
            arima_fitting_metrics = proto.Field(
                proto.MESSAGE,
                number=2,
                message="Model.ArimaFittingMetrics",
            )
            has_drift = proto.Field(
                proto.BOOL,
                number=3,
            )
            time_series_id = proto.Field(
                proto.STRING,
                number=4,
            )
            time_series_ids = proto.RepeatedField(
                proto.STRING,
                number=9,
            )
            seasonal_periods = proto.RepeatedField(
                proto.ENUM,
                number=5,
                enum="Model.SeasonalPeriod.SeasonalPeriodType",
            )
            has_holiday_effect = proto.Field(
                proto.MESSAGE,
                number=6,
                message=wrappers_pb2.BoolValue,
            )
            has_spikes_and_dips = proto.Field(
                proto.MESSAGE,
                number=7,
                message=wrappers_pb2.BoolValue,
            )
            has_step_changes = proto.Field(
                proto.MESSAGE,
                number=8,
                message=wrappers_pb2.BoolValue,
            )

        non_seasonal_order = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message="Model.ArimaOrder",
        )
        arima_fitting_metrics = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message="Model.ArimaFittingMetrics",
        )
        seasonal_periods = proto.RepeatedField(
            proto.ENUM,
            number=3,
            enum="Model.SeasonalPeriod.SeasonalPeriodType",
        )
        has_drift = proto.RepeatedField(
            proto.BOOL,
            number=4,
        )
        time_series_id = proto.RepeatedField(
            proto.STRING,
            number=5,
        )
        arima_single_model_forecasting_metrics = proto.RepeatedField(
            proto.MESSAGE,
            number=6,
            message="Model.ArimaForecastingMetrics.ArimaSingleModelForecastingMetrics",
        )

    class EvaluationMetrics(proto.Message):
        r"""Evaluation metrics of a model. These are either computed on
        all training data or just the eval data based on whether eval
        data was used during training. These are not present for
        imported models.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            regression_metrics (arrivy.google.cloud.bigquery_v2.types.Model.RegressionMetrics):
                Populated for regression models and explicit
                feedback type matrix factorization models.

                This field is a member of `oneof`_ ``metrics``.
            binary_classification_metrics (arrivy.google.cloud.bigquery_v2.types.Model.BinaryClassificationMetrics):
                Populated for binary
                classification/classifier models.

                This field is a member of `oneof`_ ``metrics``.
            multi_class_classification_metrics (arrivy.google.cloud.bigquery_v2.types.Model.MultiClassClassificationMetrics):
                Populated for multi-class
                classification/classifier models.

                This field is a member of `oneof`_ ``metrics``.
            clustering_metrics (arrivy.google.cloud.bigquery_v2.types.Model.ClusteringMetrics):
                Populated for clustering models.

                This field is a member of `oneof`_ ``metrics``.
            ranking_metrics (arrivy.google.cloud.bigquery_v2.types.Model.RankingMetrics):
                Populated for implicit feedback type matrix
                factorization models.

                This field is a member of `oneof`_ ``metrics``.
            arima_forecasting_metrics (arrivy.google.cloud.bigquery_v2.types.Model.ArimaForecastingMetrics):
                Populated for ARIMA models.

                This field is a member of `oneof`_ ``metrics``.
        """

        regression_metrics = proto.Field(
            proto.MESSAGE,
            number=1,
            oneof="metrics",
            message="Model.RegressionMetrics",
        )
        binary_classification_metrics = proto.Field(
            proto.MESSAGE,
            number=2,
            oneof="metrics",
            message="Model.BinaryClassificationMetrics",
        )
        multi_class_classification_metrics = proto.Field(
            proto.MESSAGE,
            number=3,
            oneof="metrics",
            message="Model.MultiClassClassificationMetrics",
        )
        clustering_metrics = proto.Field(
            proto.MESSAGE,
            number=4,
            oneof="metrics",
            message="Model.ClusteringMetrics",
        )
        ranking_metrics = proto.Field(
            proto.MESSAGE,
            number=5,
            oneof="metrics",
            message="Model.RankingMetrics",
        )
        arima_forecasting_metrics = proto.Field(
            proto.MESSAGE,
            number=6,
            oneof="metrics",
            message="Model.ArimaForecastingMetrics",
        )

    class DataSplitResult(proto.Message):
        r"""Data split result. This contains references to the training
        and evaluation data tables that were used to train the model.

        Attributes:
            training_table (arrivy.google.cloud.bigquery_v2.types.TableReference):
                Table reference of the training data after
                split.
            evaluation_table (arrivy.google.cloud.bigquery_v2.types.TableReference):
                Table reference of the evaluation data after
                split.
        """

        training_table = proto.Field(
            proto.MESSAGE,
            number=1,
            message=table_reference.TableReference,
        )
        evaluation_table = proto.Field(
            proto.MESSAGE,
            number=2,
            message=table_reference.TableReference,
        )

    class ArimaOrder(proto.Message):
        r"""Arima order, can be used for both non-seasonal and seasonal
        parts.

        Attributes:
            p (int):
                Order of the autoregressive part.
            d (int):
                Order of the differencing part.
            q (int):
                Order of the moving-average part.
        """

        p = proto.Field(
            proto.INT64,
            number=1,
        )
        d = proto.Field(
            proto.INT64,
            number=2,
        )
        q = proto.Field(
            proto.INT64,
            number=3,
        )

    class ArimaFittingMetrics(proto.Message):
        r"""ARIMA model fitting metrics.

        Attributes:
            log_likelihood (float):
                Log-likelihood.
            aic (float):
                AIC.
            variance (float):
                Variance.
        """

        log_likelihood = proto.Field(
            proto.DOUBLE,
            number=1,
        )
        aic = proto.Field(
            proto.DOUBLE,
            number=2,
        )
        variance = proto.Field(
            proto.DOUBLE,
            number=3,
        )

    class GlobalExplanation(proto.Message):
        r"""Global explanations containing the top most important
        features after training.

        Attributes:
            explanations (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.GlobalExplanation.Explanation]):
                A list of the top global explanations. Sorted
                by absolute value of attribution in descending
                order.
            class_label (str):
                Class label for this set of global
                explanations. Will be empty/null for binary
                logistic and linear regression models. Sorted
                alphabetically in descending order.
        """

        class Explanation(proto.Message):
            r"""Explanation for a single feature.

            Attributes:
                feature_name (str):
                    Full name of the feature. For non-numerical features, will
                    be formatted like <column_name>.<encoded_feature_name>.
                    Overall size of feature name will always be truncated to
                    first 120 characters.
                attribution (google.protobuf.wrappers_pb2.DoubleValue):
                    Attribution of feature.
            """

            feature_name = proto.Field(
                proto.STRING,
                number=1,
            )
            attribution = proto.Field(
                proto.MESSAGE,
                number=2,
                message=wrappers_pb2.DoubleValue,
            )

        explanations = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message="Model.GlobalExplanation.Explanation",
        )
        class_label = proto.Field(
            proto.STRING,
            number=2,
        )

    class TrainingRun(proto.Message):
        r"""Information about a single training query run for the model.

        Attributes:
            training_options (arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.TrainingOptions):
                Options that were used for this training run,
                includes user specified and default options that
                were used.
            start_time (google.protobuf.timestamp_pb2.Timestamp):
                The start time of this training run.
            results (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.IterationResult]):
                Output of each iteration run, results.size() <=
                max_iterations.
            evaluation_metrics (arrivy.google.cloud.bigquery_v2.types.Model.EvaluationMetrics):
                The evaluation metrics over training/eval
                data that were computed at the end of training.
            data_split_result (arrivy.google.cloud.bigquery_v2.types.Model.DataSplitResult):
                Data split result of the training run. Only
                set when the input data is actually split.
            global_explanations (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.GlobalExplanation]):
                Global explanations for important features of
                the model. For multi-class models, there is one
                entry for each label class. For other models,
                there is only one entry in the list.
        """

        class TrainingOptions(proto.Message):
            r"""Options used in model training.

            Attributes:
                max_iterations (int):
                    The maximum number of iterations in training.
                    Used only for iterative training algorithms.
                loss_type (arrivy.google.cloud.bigquery_v2.types.Model.LossType):
                    Type of loss function used during training
                    run.
                learn_rate (float):
                    Learning rate in training. Used only for
                    iterative training algorithms.
                l1_regularization (google.protobuf.wrappers_pb2.DoubleValue):
                    L1 regularization coefficient.
                l2_regularization (google.protobuf.wrappers_pb2.DoubleValue):
                    L2 regularization coefficient.
                min_relative_progress (google.protobuf.wrappers_pb2.DoubleValue):
                    When early_stop is true, stops training when accuracy
                    improvement is less than 'min_relative_progress'. Used only
                    for iterative training algorithms.
                warm_start (google.protobuf.wrappers_pb2.BoolValue):
                    Whether to train a model from the last
                    checkpoint.
                early_stop (google.protobuf.wrappers_pb2.BoolValue):
                    Whether to stop early when the loss doesn't improve
                    significantly any more (compared to min_relative_progress).
                    Used only for iterative training algorithms.
                input_label_columns (Sequence[str]):
                    Name of input label columns in training data.
                data_split_method (arrivy.google.cloud.bigquery_v2.types.Model.DataSplitMethod):
                    The data split type for training and
                    evaluation, e.g. RANDOM.
                data_split_eval_fraction (float):
                    The fraction of evaluation data over the
                    whole input data. The rest of data will be used
                    as training data. The format should be double.
                    Accurate to two decimal places.
                    Default value is 0.2.
                data_split_column (str):
                    The column to split data with. This column won't be used as
                    a feature.

                    1. When data_split_method is CUSTOM, the corresponding
                       column should be boolean. The rows with true value tag
                       are eval data, and the false are training data.
                    2. When data_split_method is SEQ, the first
                       DATA_SPLIT_EVAL_FRACTION rows (from smallest to largest)
                       in the corresponding column are used as training data,
                       and the rest are eval data. It respects the order in
                       Orderable data types:
                       https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#data-type-properties
                learn_rate_strategy (arrivy.google.cloud.bigquery_v2.types.Model.LearnRateStrategy):
                    The strategy to determine learn rate for the
                    current iteration.
                initial_learn_rate (float):
                    Specifies the initial learning rate for the
                    line search learn rate strategy.
                label_class_weights (Mapping[str, float]):
                    Weights associated with each label class, for
                    rebalancing the training data. Only applicable
                    for classification models.
                user_column (str):
                    User column specified for matrix
                    factorization models.
                item_column (str):
                    Item column specified for matrix
                    factorization models.
                distance_type (arrivy.google.cloud.bigquery_v2.types.Model.DistanceType):
                    Distance type for clustering models.
                num_clusters (int):
                    Number of clusters for clustering models.
                model_uri (str):
                    Google Cloud Storage URI from which the model
                    was imported. Only applicable for imported
                    models.
                optimization_strategy (arrivy.google.cloud.bigquery_v2.types.Model.OptimizationStrategy):
                    Optimization strategy for training linear
                    regression models.
                hidden_units (Sequence[int]):
                    Hidden units for dnn models.
                batch_size (int):
                    Batch size for dnn models.
                dropout (google.protobuf.wrappers_pb2.DoubleValue):
                    Dropout probability for dnn models.
                max_tree_depth (int):
                    Maximum depth of a tree for boosted tree
                    models.
                subsample (float):
                    Subsample fraction of the training data to
                    grow tree to prevent overfitting for boosted
                    tree models.
                min_split_loss (google.protobuf.wrappers_pb2.DoubleValue):
                    Minimum split loss for boosted tree models.
                num_factors (int):
                    Num factors specified for matrix
                    factorization models.
                feedback_type (arrivy.google.cloud.bigquery_v2.types.Model.FeedbackType):
                    Feedback type that specifies which algorithm
                    to run for matrix factorization.
                wals_alpha (google.protobuf.wrappers_pb2.DoubleValue):
                    Hyperparameter for matrix factoration when
                    implicit feedback type is specified.
                kmeans_initialization_method (arrivy.google.cloud.bigquery_v2.types.Model.KmeansEnums.KmeansInitializationMethod):
                    The method used to initialize the centroids
                    for kmeans algorithm.
                kmeans_initialization_column (str):
                    The column used to provide the initial centroids for kmeans
                    algorithm when kmeans_initialization_method is CUSTOM.
                time_series_timestamp_column (str):
                    Column to be designated as time series
                    timestamp for ARIMA model.
                time_series_data_column (str):
                    Column to be designated as time series data
                    for ARIMA model.
                auto_arima (bool):
                    Whether to enable auto ARIMA or not.
                non_seasonal_order (arrivy.google.cloud.bigquery_v2.types.Model.ArimaOrder):
                    A specification of the non-seasonal part of
                    the ARIMA model: the three components (p, d, q)
                    are the AR order, the degree of differencing,
                    and the MA order.
                data_frequency (arrivy.google.cloud.bigquery_v2.types.Model.DataFrequency):
                    The data frequency of a time series.
                include_drift (bool):
                    Include drift when fitting an ARIMA model.
                holiday_region (arrivy.google.cloud.bigquery_v2.types.Model.HolidayRegion):
                    The geographical region based on which the
                    holidays are considered in time series modeling.
                    If a valid value is specified, then holiday
                    effects modeling is enabled.
                time_series_id_column (str):
                    The time series id column that was used
                    during ARIMA model training.
                time_series_id_columns (Sequence[str]):
                    The time series id columns that were used
                    during ARIMA model training.
                horizon (int):
                    The number of periods ahead that need to be
                    forecasted.
                preserve_input_structs (bool):
                    Whether to preserve the input structs in output feature
                    names. Suppose there is a struct A with field b. When false
                    (default), the output feature name is A_b. When true, the
                    output feature name is A.b.
                auto_arima_max_order (int):
                    The max value of non-seasonal p and q.
                decompose_time_series (google.protobuf.wrappers_pb2.BoolValue):
                    If true, perform decompose time series and
                    save the results.
                clean_spikes_and_dips (google.protobuf.wrappers_pb2.BoolValue):
                    If true, clean spikes and dips in the input
                    time series.
                adjust_step_changes (google.protobuf.wrappers_pb2.BoolValue):
                    If true, detect step changes and make data
                    adjustment in the input time series.
            """

            max_iterations = proto.Field(
                proto.INT64,
                number=1,
            )
            loss_type = proto.Field(
                proto.ENUM,
                number=2,
                enum="Model.LossType",
            )
            learn_rate = proto.Field(
                proto.DOUBLE,
                number=3,
            )
            l1_regularization = proto.Field(
                proto.MESSAGE,
                number=4,
                message=wrappers_pb2.DoubleValue,
            )
            l2_regularization = proto.Field(
                proto.MESSAGE,
                number=5,
                message=wrappers_pb2.DoubleValue,
            )
            min_relative_progress = proto.Field(
                proto.MESSAGE,
                number=6,
                message=wrappers_pb2.DoubleValue,
            )
            warm_start = proto.Field(
                proto.MESSAGE,
                number=7,
                message=wrappers_pb2.BoolValue,
            )
            early_stop = proto.Field(
                proto.MESSAGE,
                number=8,
                message=wrappers_pb2.BoolValue,
            )
            input_label_columns = proto.RepeatedField(
                proto.STRING,
                number=9,
            )
            data_split_method = proto.Field(
                proto.ENUM,
                number=10,
                enum="Model.DataSplitMethod",
            )
            data_split_eval_fraction = proto.Field(
                proto.DOUBLE,
                number=11,
            )
            data_split_column = proto.Field(
                proto.STRING,
                number=12,
            )
            learn_rate_strategy = proto.Field(
                proto.ENUM,
                number=13,
                enum="Model.LearnRateStrategy",
            )
            initial_learn_rate = proto.Field(
                proto.DOUBLE,
                number=16,
            )
            label_class_weights = proto.MapField(
                proto.STRING,
                proto.DOUBLE,
                number=17,
            )
            user_column = proto.Field(
                proto.STRING,
                number=18,
            )
            item_column = proto.Field(
                proto.STRING,
                number=19,
            )
            distance_type = proto.Field(
                proto.ENUM,
                number=20,
                enum="Model.DistanceType",
            )
            num_clusters = proto.Field(
                proto.INT64,
                number=21,
            )
            model_uri = proto.Field(
                proto.STRING,
                number=22,
            )
            optimization_strategy = proto.Field(
                proto.ENUM,
                number=23,
                enum="Model.OptimizationStrategy",
            )
            hidden_units = proto.RepeatedField(
                proto.INT64,
                number=24,
            )
            batch_size = proto.Field(
                proto.INT64,
                number=25,
            )
            dropout = proto.Field(
                proto.MESSAGE,
                number=26,
                message=wrappers_pb2.DoubleValue,
            )
            max_tree_depth = proto.Field(
                proto.INT64,
                number=27,
            )
            subsample = proto.Field(
                proto.DOUBLE,
                number=28,
            )
            min_split_loss = proto.Field(
                proto.MESSAGE,
                number=29,
                message=wrappers_pb2.DoubleValue,
            )
            num_factors = proto.Field(
                proto.INT64,
                number=30,
            )
            feedback_type = proto.Field(
                proto.ENUM,
                number=31,
                enum="Model.FeedbackType",
            )
            wals_alpha = proto.Field(
                proto.MESSAGE,
                number=32,
                message=wrappers_pb2.DoubleValue,
            )
            kmeans_initialization_method = proto.Field(
                proto.ENUM,
                number=33,
                enum="Model.KmeansEnums.KmeansInitializationMethod",
            )
            kmeans_initialization_column = proto.Field(
                proto.STRING,
                number=34,
            )
            time_series_timestamp_column = proto.Field(
                proto.STRING,
                number=35,
            )
            time_series_data_column = proto.Field(
                proto.STRING,
                number=36,
            )
            auto_arima = proto.Field(
                proto.BOOL,
                number=37,
            )
            non_seasonal_order = proto.Field(
                proto.MESSAGE,
                number=38,
                message="Model.ArimaOrder",
            )
            data_frequency = proto.Field(
                proto.ENUM,
                number=39,
                enum="Model.DataFrequency",
            )
            include_drift = proto.Field(
                proto.BOOL,
                number=41,
            )
            holiday_region = proto.Field(
                proto.ENUM,
                number=42,
                enum="Model.HolidayRegion",
            )
            time_series_id_column = proto.Field(
                proto.STRING,
                number=43,
            )
            time_series_id_columns = proto.RepeatedField(
                proto.STRING,
                number=51,
            )
            horizon = proto.Field(
                proto.INT64,
                number=44,
            )
            preserve_input_structs = proto.Field(
                proto.BOOL,
                number=45,
            )
            auto_arima_max_order = proto.Field(
                proto.INT64,
                number=46,
            )
            decompose_time_series = proto.Field(
                proto.MESSAGE,
                number=50,
                message=wrappers_pb2.BoolValue,
            )
            clean_spikes_and_dips = proto.Field(
                proto.MESSAGE,
                number=52,
                message=wrappers_pb2.BoolValue,
            )
            adjust_step_changes = proto.Field(
                proto.MESSAGE,
                number=53,
                message=wrappers_pb2.BoolValue,
            )

        class IterationResult(proto.Message):
            r"""Information about a single iteration of the training run.

            Attributes:
                index (google.protobuf.wrappers_pb2.Int32Value):
                    Index of the iteration, 0 based.
                duration_ms (google.protobuf.wrappers_pb2.Int64Value):
                    Time taken to run the iteration in
                    milliseconds.
                training_loss (google.protobuf.wrappers_pb2.DoubleValue):
                    Loss computed on the training data at the end
                    of iteration.
                eval_loss (google.protobuf.wrappers_pb2.DoubleValue):
                    Loss computed on the eval data at the end of
                    iteration.
                learn_rate (float):
                    Learn rate used for this iteration.
                cluster_infos (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.IterationResult.ClusterInfo]):
                    Information about top clusters for clustering
                    models.
                arima_result (arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.IterationResult.ArimaResult):

            """

            class ClusterInfo(proto.Message):
                r"""Information about a single cluster for clustering model.

                Attributes:
                    centroid_id (int):
                        Centroid id.
                    cluster_radius (google.protobuf.wrappers_pb2.DoubleValue):
                        Cluster radius, the average distance from
                        centroid to each point assigned to the cluster.
                    cluster_size (google.protobuf.wrappers_pb2.Int64Value):
                        Cluster size, the total number of points
                        assigned to the cluster.
                """

                centroid_id = proto.Field(
                    proto.INT64,
                    number=1,
                )
                cluster_radius = proto.Field(
                    proto.MESSAGE,
                    number=2,
                    message=wrappers_pb2.DoubleValue,
                )
                cluster_size = proto.Field(
                    proto.MESSAGE,
                    number=3,
                    message=wrappers_pb2.Int64Value,
                )

            class ArimaResult(proto.Message):
                r"""(Auto-)arima fitting result. Wrap everything in ArimaResult
                for easier refactoring if we want to use model-specific
                iteration results.

                Attributes:
                    arima_model_info (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.IterationResult.ArimaResult.ArimaModelInfo]):
                        This message is repeated because there are
                        multiple arima models fitted in auto-arima. For
                        non-auto-arima model, its size is one.
                    seasonal_periods (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.SeasonalPeriod.SeasonalPeriodType]):
                        Seasonal periods. Repeated because multiple
                        periods are supported for one time series.
                """

                class ArimaCoefficients(proto.Message):
                    r"""Arima coefficients.

                    Attributes:
                        auto_regressive_coefficients (Sequence[float]):
                            Auto-regressive coefficients, an array of
                            double.
                        moving_average_coefficients (Sequence[float]):
                            Moving-average coefficients, an array of
                            double.
                        intercept_coefficient (float):
                            Intercept coefficient, just a double not an
                            array.
                    """

                    auto_regressive_coefficients = proto.RepeatedField(
                        proto.DOUBLE,
                        number=1,
                    )
                    moving_average_coefficients = proto.RepeatedField(
                        proto.DOUBLE,
                        number=2,
                    )
                    intercept_coefficient = proto.Field(
                        proto.DOUBLE,
                        number=3,
                    )

                class ArimaModelInfo(proto.Message):
                    r"""Arima model information.

                    Attributes:
                        non_seasonal_order (arrivy.google.cloud.bigquery_v2.types.Model.ArimaOrder):
                            Non-seasonal order.
                        arima_coefficients (arrivy.google.cloud.bigquery_v2.types.Model.TrainingRun.IterationResult.ArimaResult.ArimaCoefficients):
                            Arima coefficients.
                        arima_fitting_metrics (arrivy.google.cloud.bigquery_v2.types.Model.ArimaFittingMetrics):
                            Arima fitting metrics.
                        has_drift (bool):
                            Whether Arima model fitted with drift or not.
                            It is always false when d is not 1.
                        time_series_id (str):
                            The time_series_id value for this time series. It will be
                            one of the unique values from the time_series_id_column
                            specified during ARIMA model training. Only present when
                            time_series_id_column training option was used.
                        time_series_ids (Sequence[str]):
                            The tuple of time_series_ids identifying this time series.
                            It will be one of the unique tuples of values present in the
                            time_series_id_columns specified during ARIMA model
                            training. Only present when time_series_id_columns training
                            option was used and the order of values here are same as the
                            order of time_series_id_columns.
                        seasonal_periods (Sequence[arrivy.google.cloud.bigquery_v2.types.Model.SeasonalPeriod.SeasonalPeriodType]):
                            Seasonal periods. Repeated because multiple
                            periods are supported for one time series.
                        has_holiday_effect (google.protobuf.wrappers_pb2.BoolValue):
                            If true, holiday_effect is a part of time series
                            decomposition result.
                        has_spikes_and_dips (google.protobuf.wrappers_pb2.BoolValue):
                            If true, spikes_and_dips is a part of time series
                            decomposition result.
                        has_step_changes (google.protobuf.wrappers_pb2.BoolValue):
                            If true, step_changes is a part of time series decomposition
                            result.
                    """

                    non_seasonal_order = proto.Field(
                        proto.MESSAGE,
                        number=1,
                        message="Model.ArimaOrder",
                    )
                    arima_coefficients = proto.Field(
                        proto.MESSAGE,
                        number=2,
                        message="Model.TrainingRun.IterationResult.ArimaResult.ArimaCoefficients",
                    )
                    arima_fitting_metrics = proto.Field(
                        proto.MESSAGE,
                        number=3,
                        message="Model.ArimaFittingMetrics",
                    )
                    has_drift = proto.Field(
                        proto.BOOL,
                        number=4,
                    )
                    time_series_id = proto.Field(
                        proto.STRING,
                        number=5,
                    )
                    time_series_ids = proto.RepeatedField(
                        proto.STRING,
                        number=10,
                    )
                    seasonal_periods = proto.RepeatedField(
                        proto.ENUM,
                        number=6,
                        enum="Model.SeasonalPeriod.SeasonalPeriodType",
                    )
                    has_holiday_effect = proto.Field(
                        proto.MESSAGE,
                        number=7,
                        message=wrappers_pb2.BoolValue,
                    )
                    has_spikes_and_dips = proto.Field(
                        proto.MESSAGE,
                        number=8,
                        message=wrappers_pb2.BoolValue,
                    )
                    has_step_changes = proto.Field(
                        proto.MESSAGE,
                        number=9,
                        message=wrappers_pb2.BoolValue,
                    )

                arima_model_info = proto.RepeatedField(
                    proto.MESSAGE,
                    number=1,
                    message="Model.TrainingRun.IterationResult.ArimaResult.ArimaModelInfo",
                )
                seasonal_periods = proto.RepeatedField(
                    proto.ENUM,
                    number=2,
                    enum="Model.SeasonalPeriod.SeasonalPeriodType",
                )

            index = proto.Field(
                proto.MESSAGE,
                number=1,
                message=wrappers_pb2.Int32Value,
            )
            duration_ms = proto.Field(
                proto.MESSAGE,
                number=4,
                message=wrappers_pb2.Int64Value,
            )
            training_loss = proto.Field(
                proto.MESSAGE,
                number=5,
                message=wrappers_pb2.DoubleValue,
            )
            eval_loss = proto.Field(
                proto.MESSAGE,
                number=6,
                message=wrappers_pb2.DoubleValue,
            )
            learn_rate = proto.Field(
                proto.DOUBLE,
                number=7,
            )
            cluster_infos = proto.RepeatedField(
                proto.MESSAGE,
                number=8,
                message="Model.TrainingRun.IterationResult.ClusterInfo",
            )
            arima_result = proto.Field(
                proto.MESSAGE,
                number=9,
                message="Model.TrainingRun.IterationResult.ArimaResult",
            )

        training_options = proto.Field(
            proto.MESSAGE,
            number=1,
            message="Model.TrainingRun.TrainingOptions",
        )
        start_time = proto.Field(
            proto.MESSAGE,
            number=8,
            message=timestamp_pb2.Timestamp,
        )
        results = proto.RepeatedField(
            proto.MESSAGE,
            number=6,
            message="Model.TrainingRun.IterationResult",
        )
        evaluation_metrics = proto.Field(
            proto.MESSAGE,
            number=7,
            message="Model.EvaluationMetrics",
        )
        data_split_result = proto.Field(
            proto.MESSAGE,
            number=9,
            message="Model.DataSplitResult",
        )
        global_explanations = proto.RepeatedField(
            proto.MESSAGE,
            number=10,
            message="Model.GlobalExplanation",
        )

    etag = proto.Field(
        proto.STRING,
        number=1,
    )
    model_reference = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gcb_model_reference.ModelReference,
    )
    creation_time = proto.Field(
        proto.INT64,
        number=5,
    )
    last_modified_time = proto.Field(
        proto.INT64,
        number=6,
    )
    description = proto.Field(
        proto.STRING,
        number=12,
    )
    friendly_name = proto.Field(
        proto.STRING,
        number=14,
    )
    labels = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=15,
    )
    expiration_time = proto.Field(
        proto.INT64,
        number=16,
    )
    location = proto.Field(
        proto.STRING,
        number=13,
    )
    encryption_configuration = proto.Field(
        proto.MESSAGE,
        number=17,
        message=encryption_config.EncryptionConfiguration,
    )
    model_type = proto.Field(
        proto.ENUM,
        number=7,
        enum=ModelType,
    )
    training_runs = proto.RepeatedField(
        proto.MESSAGE,
        number=9,
        message=TrainingRun,
    )
    feature_columns = proto.RepeatedField(
        proto.MESSAGE,
        number=10,
        message=standard_sql.StandardSqlField,
    )
    label_columns = proto.RepeatedField(
        proto.MESSAGE,
        number=11,
        message=standard_sql.StandardSqlField,
    )
    best_trial_id = proto.Field(
        proto.INT64,
        number=19,
    )


class GetModelRequest(proto.Message):
    r"""

    Attributes:
        project_id (str):
            Required. Project ID of the requested model.
        dataset_id (str):
            Required. Dataset ID of the requested model.
        model_id (str):
            Required. Model ID of the requested model.
    """

    project_id = proto.Field(
        proto.STRING,
        number=1,
    )
    dataset_id = proto.Field(
        proto.STRING,
        number=2,
    )
    model_id = proto.Field(
        proto.STRING,
        number=3,
    )


class PatchModelRequest(proto.Message):
    r"""

    Attributes:
        project_id (str):
            Required. Project ID of the model to patch.
        dataset_id (str):
            Required. Dataset ID of the model to patch.
        model_id (str):
            Required. Model ID of the model to patch.
        model (arrivy.google.cloud.bigquery_v2.types.Model):
            Required. Patched model.
            Follows RFC5789 patch semantics. Missing fields
            are not updated. To clear a field, explicitly
            set to default value.
    """

    project_id = proto.Field(
        proto.STRING,
        number=1,
    )
    dataset_id = proto.Field(
        proto.STRING,
        number=2,
    )
    model_id = proto.Field(
        proto.STRING,
        number=3,
    )
    model = proto.Field(
        proto.MESSAGE,
        number=4,
        message="Model",
    )


class DeleteModelRequest(proto.Message):
    r"""

    Attributes:
        project_id (str):
            Required. Project ID of the model to delete.
        dataset_id (str):
            Required. Dataset ID of the model to delete.
        model_id (str):
            Required. Model ID of the model to delete.
    """

    project_id = proto.Field(
        proto.STRING,
        number=1,
    )
    dataset_id = proto.Field(
        proto.STRING,
        number=2,
    )
    model_id = proto.Field(
        proto.STRING,
        number=3,
    )


class ListModelsRequest(proto.Message):
    r"""

    Attributes:
        project_id (str):
            Required. Project ID of the models to list.
        dataset_id (str):
            Required. Dataset ID of the models to list.
        max_results (google.protobuf.wrappers_pb2.UInt32Value):
            The maximum number of results to return in a
            single response page. Leverage the page tokens
            to iterate through the entire collection.
        page_token (str):
            Page token, returned by a previous call to
            request the next page of results
    """

    project_id = proto.Field(
        proto.STRING,
        number=1,
    )
    dataset_id = proto.Field(
        proto.STRING,
        number=2,
    )
    max_results = proto.Field(
        proto.MESSAGE,
        number=3,
        message=wrappers_pb2.UInt32Value,
    )
    page_token = proto.Field(
        proto.STRING,
        number=4,
    )


class ListModelsResponse(proto.Message):
    r"""

    Attributes:
        models (Sequence[arrivy.google.cloud.bigquery_v2.types.Model]):
            Models in the requested dataset. Only the following fields
            are populated: model_reference, model_type, creation_time,
            last_modified_time and labels.
        next_page_token (str):
            A token to request the next page of results.
    """

    @property
    def raw_page(self):
        return self

    models = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message="Model",
    )
    next_page_token = proto.Field(
        proto.STRING,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""System tests for pandas connector."""

import collections
import datetime
import decimal
import json
import io
import operator
import warnings

import google.api_core.retry
import pytest

try:
    import importlib.metadata as metadata
except ImportError:
    import importlib_metadata as metadata

from arrivy.google.cloud import bigquery

from arrivy.google.cloud.bigquery import enums

from . import helpers


pandas = pytest.importorskip("pandas", minversion="0.23.0")
pyarrow = pytest.importorskip("pyarrow")
numpy = pytest.importorskip("numpy")

bigquery_storage = pytest.importorskip(
    "arrivy.google.cloud.bigquery_storage", minversion="2.0.0"
)

if pandas is not None:
    PANDAS_INSTALLED_VERSION = metadata.version("pandas")
else:
    PANDAS_INSTALLED_VERSION = "0.0.0"


class MissingDataError(Exception):
    pass


def test_load_table_from_dataframe_w_automatic_schema(bigquery_client, dataset_id):
    """Test that a DataFrame with dtypes that map well to BigQuery types
    can be uploaded without specifying a schema.

    https://github.com/googleapis/google-cloud-python/issues/9044
    """
    df_data = collections.OrderedDict(
        [
            ("bool_col", pandas.Series([True, False, True], dtype="bool")),
            (
                "ts_col",
                pandas.Series(
                    [
                        datetime.datetime(2010, 1, 2, 3, 44, 50),
                        datetime.datetime(2011, 2, 3, 14, 50, 59),
                        datetime.datetime(2012, 3, 14, 15, 16),
                    ],
                    dtype="datetime64[ns]",
                ).dt.tz_localize(datetime.timezone.utc),
            ),
            (
                "dt_col_no_tz",
                pandas.Series(
                    [
                        datetime.datetime(2010, 1, 2, 3, 44, 50),
                        datetime.datetime(2011, 2, 3, 14, 50, 59),
                        datetime.datetime(2012, 3, 14, 15, 16),
                    ],
                    dtype="datetime64[ns]",
                ),
            ),
            ("float32_col", pandas.Series([1.0, 2.0, 3.0], dtype="float32")),
            ("float64_col", pandas.Series([4.0, 5.0, 6.0], dtype="float64")),
            ("int8_col", pandas.Series([-12, -11, -10], dtype="int8")),
            ("int16_col", pandas.Series([-9, -8, -7], dtype="int16")),
            ("int32_col", pandas.Series([-6, -5, -4], dtype="int32")),
            ("int64_col", pandas.Series([-3, -2, -1], dtype="int64")),
            ("uint8_col", pandas.Series([0, 1, 2], dtype="uint8")),
            ("uint16_col", pandas.Series([3, 4, 5], dtype="uint16")),
            ("uint32_col", pandas.Series([6, 7, 8], dtype="uint32")),
            (
                "date_col",
                pandas.Series(
                    [
                        datetime.date(2010, 1, 2),
                        datetime.date(2011, 2, 3),
                        datetime.date(2012, 3, 14),
                    ],
                    dtype="dbdate",
                ),
            ),
            (
                "time_col",
                pandas.Series(
                    [
                        datetime.time(3, 44, 50),
                        datetime.time(14, 50, 59),
                        datetime.time(15, 16),
                    ],
                    dtype="dbtime",
                ),
            ),
            ("array_bool_col", pandas.Series([[True], [False], [True]])),
            (
                "array_ts_col",
                pandas.Series(
                    [
                        [
                            datetime.datetime(
                                2010, 1, 2, 3, 44, 50, tzinfo=datetime.timezone.utc
                            ),
                        ],
                        [
                            datetime.datetime(
                                2011, 2, 3, 14, 50, 59, tzinfo=datetime.timezone.utc
                            ),
                        ],
                        [
                            datetime.datetime(
                                2012, 3, 14, 15, 16, tzinfo=datetime.timezone.utc
                            ),
                        ],
                    ],
                ),
            ),
            (
                "array_dt_col_no_tz",
                pandas.Series(
                    [
                        [datetime.datetime(2010, 1, 2, 3, 44, 50)],
                        [datetime.datetime(2011, 2, 3, 14, 50, 59)],
                        [datetime.datetime(2012, 3, 14, 15, 16)],
                    ],
                ),
            ),
            (
                "array_float32_col",
                pandas.Series(
                    [numpy.array([_], dtype="float32") for _ in [1.0, 2.0, 3.0]]
                ),
            ),
            (
                "array_float64_col",
                pandas.Series(
                    [numpy.array([_], dtype="float64") for _ in [4.0, 5.0, 6.0]]
                ),
            ),
            (
                "array_int8_col",
                pandas.Series(
                    [numpy.array([_], dtype="int8") for _ in [-12, -11, -10]]
                ),
            ),
            (
                "array_int16_col",
                pandas.Series([numpy.array([_], dtype="int16") for _ in [-9, -8, -7]]),
            ),
            (
                "array_int32_col",
                pandas.Series([numpy.array([_], dtype="int32") for _ in [-6, -5, -4]]),
            ),
            (
                "array_int64_col",
                pandas.Series([numpy.array([_], dtype="int64") for _ in [-3, -2, -1]]),
            ),
            (
                "array_uint8_col",
                pandas.Series([numpy.array([_], dtype="uint8") for _ in [0, 1, 2]]),
            ),
            (
                "array_uint16_col",
                pandas.Series([numpy.array([_], dtype="uint16") for _ in [3, 4, 5]]),
            ),
            (
                "array_uint32_col",
                pandas.Series([numpy.array([_], dtype="uint32") for _ in [6, 7, 8]]),
            ),
        ]
    )
    dataframe = pandas.DataFrame(df_data, columns=df_data.keys())

    table_id = "{}.{}.load_table_from_dataframe_w_automatic_schema".format(
        bigquery_client.project, dataset_id
    )

    load_job = bigquery_client.load_table_from_dataframe(dataframe, table_id)
    load_job.result()

    table = bigquery_client.get_table(table_id)
    assert tuple(table.schema) == (
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("ts_col", "TIMESTAMP"),
        bigquery.SchemaField("dt_col_no_tz", "DATETIME"),
        bigquery.SchemaField("float32_col", "FLOAT"),
        bigquery.SchemaField("float64_col", "FLOAT"),
        bigquery.SchemaField("int8_col", "INTEGER"),
        bigquery.SchemaField("int16_col", "INTEGER"),
        bigquery.SchemaField("int32_col", "INTEGER"),
        bigquery.SchemaField("int64_col", "INTEGER"),
        bigquery.SchemaField("uint8_col", "INTEGER"),
        bigquery.SchemaField("uint16_col", "INTEGER"),
        bigquery.SchemaField("uint32_col", "INTEGER"),
        bigquery.SchemaField("date_col", "DATE"),
        bigquery.SchemaField("time_col", "TIME"),
        bigquery.SchemaField("array_bool_col", "BOOLEAN", mode="REPEATED"),
        bigquery.SchemaField("array_ts_col", "TIMESTAMP", mode="REPEATED"),
        bigquery.SchemaField("array_dt_col_no_tz", "DATETIME", mode="REPEATED"),
        bigquery.SchemaField("array_float32_col", "FLOAT", mode="REPEATED"),
        bigquery.SchemaField("array_float64_col", "FLOAT", mode="REPEATED"),
        bigquery.SchemaField("array_int8_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_int16_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_int32_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_int64_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_uint8_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_uint16_col", "INTEGER", mode="REPEATED"),
        bigquery.SchemaField("array_uint32_col", "INTEGER", mode="REPEATED"),
    )

    assert numpy.array(
        sorted(map(list, bigquery_client.list_rows(table)), key=lambda r: r[5]),
        dtype="object",
    ).transpose().tolist() == [
        # bool_col
        [True, False, True],
        # ts_col
        [
            datetime.datetime(2010, 1, 2, 3, 44, 50, tzinfo=datetime.timezone.utc),
            datetime.datetime(2011, 2, 3, 14, 50, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(2012, 3, 14, 15, 16, tzinfo=datetime.timezone.utc),
        ],
        # dt_col_no_tz
        [
            datetime.datetime(2010, 1, 2, 3, 44, 50),
            datetime.datetime(2011, 2, 3, 14, 50, 59),
            datetime.datetime(2012, 3, 14, 15, 16),
        ],
        # float32_col
        [1.0, 2.0, 3.0],
        # float64_col
        [4.0, 5.0, 6.0],
        # int8_col
        [-12, -11, -10],
        # int16_col
        [-9, -8, -7],
        # int32_col
        [-6, -5, -4],
        # int64_col
        [-3, -2, -1],
        # uint8_col
        [0, 1, 2],
        # uint16_col
        [3, 4, 5],
        # uint32_col
        [6, 7, 8],
        # date_col
        [
            datetime.date(2010, 1, 2),
            datetime.date(2011, 2, 3),
            datetime.date(2012, 3, 14),
        ],
        # time_col
        [datetime.time(3, 44, 50), datetime.time(14, 50, 59), datetime.time(15, 16)],
        # array_bool_col
        [[True], [False], [True]],
        # array_ts_col
        [
            [datetime.datetime(2010, 1, 2, 3, 44, 50, tzinfo=datetime.timezone.utc)],
            [datetime.datetime(2011, 2, 3, 14, 50, 59, tzinfo=datetime.timezone.utc)],
            [datetime.datetime(2012, 3, 14, 15, 16, tzinfo=datetime.timezone.utc)],
        ],
        # array_dt_col
        [
            [datetime.datetime(2010, 1, 2, 3, 44, 50)],
            [datetime.datetime(2011, 2, 3, 14, 50, 59)],
            [datetime.datetime(2012, 3, 14, 15, 16)],
        ],
        # array_float32_col
        [[1.0], [2.0], [3.0]],
        # array_float64_col
        [[4.0], [5.0], [6.0]],
        # array_int8_col
        [[-12], [-11], [-10]],
        # array_int16_col
        [[-9], [-8], [-7]],
        # array_int32_col
        [[-6], [-5], [-4]],
        # array_int64_col
        [[-3], [-2], [-1]],
        # array_uint8_col
        [[0], [1], [2]],
        # array_uint16_col
        [[3], [4], [5]],
        # array_uint32_col
        [[6], [7], [8]],
    ]


@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
def test_load_table_from_dataframe_w_nullable_int64_datatype(
    bigquery_client, dataset_id
):
    """Test that a DataFrame containing column with None-type values and int64 datatype
    can be uploaded if a BigQuery schema is specified.

    https://github.com/googleapis/python-bigquery/issues/22
    """
    table_id = "{}.{}.load_table_from_dataframe_w_nullable_int64_datatype".format(
        bigquery_client.project, dataset_id
    )
    table_schema = (bigquery.SchemaField("x", "INTEGER", mode="NULLABLE"),)
    table = helpers.retry_403(bigquery_client.create_table)(
        bigquery.Table(table_id, schema=table_schema)
    )

    df_data = collections.OrderedDict(
        [("x", pandas.Series([1, 2, None, 4], dtype="Int64"))]
    )
    dataframe = pandas.DataFrame(df_data, columns=df_data.keys())
    load_job = bigquery_client.load_table_from_dataframe(dataframe, table_id)
    load_job.result()
    table = bigquery_client.get_table(table_id)
    assert tuple(table.schema) == (bigquery.SchemaField("x", "INTEGER"),)
    assert table.num_rows == 4


@pytest.mark.skipif(
    PANDAS_INSTALLED_VERSION[0:2].startswith("0."),
    reason="Only `pandas version >=1.0.0` is supported",
)
def test_load_table_from_dataframe_w_nullable_int64_datatype_automatic_schema(
    bigquery_client, dataset_id, table_id
):
    """Test that a DataFrame containing column with None-type values and int64 datatype
    can be uploaded without specifying a schema.

    https://github.com/googleapis/python-bigquery/issues/22
    """

    df_data = collections.OrderedDict(
        [("x", pandas.Series([1, 2, None, 4], dtype="Int64"))]
    )
    dataframe = pandas.DataFrame(df_data, columns=df_data.keys())
    load_job = bigquery_client.load_table_from_dataframe(dataframe, table_id)
    load_job.result()
    table = bigquery_client.get_table(table_id)
    assert tuple(table.schema) == (bigquery.SchemaField("x", "INTEGER"),)
    assert table.num_rows == 4


def test_load_table_from_dataframe_w_nulls(bigquery_client, dataset_id):
    """Test that a DataFrame with null columns can be uploaded if a
    BigQuery schema is specified.

    See: https://github.com/googleapis/google-cloud-python/issues/7370
    """
    # Schema with all scalar types.
    table_schema = (
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("bytes_col", "BYTES"),
        bigquery.SchemaField("date_col", "DATE"),
        bigquery.SchemaField("dt_col", "DATETIME"),
        bigquery.SchemaField("float_col", "FLOAT"),
        bigquery.SchemaField("geo_col", "GEOGRAPHY"),
        bigquery.SchemaField("int_col", "INTEGER"),
        bigquery.SchemaField("num_col", "NUMERIC"),
        bigquery.SchemaField("str_col", "STRING"),
        bigquery.SchemaField("time_col", "TIME"),
        bigquery.SchemaField("ts_col", "TIMESTAMP"),
        bigquery.SchemaField("bignum_col", "BIGNUMERIC"),
    )

    num_rows = 100
    nulls = [None] * num_rows
    df_data = [
        ("bool_col", nulls),
        ("bytes_col", nulls),
        ("date_col", nulls),
        ("dt_col", nulls),
        ("float_col", nulls),
        ("geo_col", nulls),
        ("int_col", nulls),
        ("num_col", nulls),
        ("str_col", nulls),
        ("time_col", nulls),
        ("ts_col", nulls),
        ("bignum_col", nulls),
    ]
    df_data = collections.OrderedDict(df_data)
    dataframe = pandas.DataFrame(df_data, columns=df_data.keys())

    table_id = "{}.{}.load_table_from_dataframe_w_nulls".format(
        bigquery_client.project, dataset_id
    )

    # Create the table before loading so that schema mismatch errors are
    # identified.
    table = helpers.retry_403(bigquery_client.create_table)(
        bigquery.Table(table_id, schema=table_schema)
    )

    job_config = bigquery.LoadJobConfig(schema=table_schema)
    load_job = bigquery_client.load_table_from_dataframe(
        dataframe, table_id, job_config=job_config
    )
    load_job.result()

    table = bigquery_client.get_table(table)
    assert tuple(table.schema) == table_schema
    assert table.num_rows == num_rows


def test_load_table_from_dataframe_w_required(bigquery_client, dataset_id):
    """Test that a DataFrame can be uploaded to a table with required columns.

    See: https://github.com/googleapis/google-cloud-python/issues/8093
    """
    table_schema = (
        bigquery.SchemaField("name", "STRING", mode="REQUIRED"),
        bigquery.SchemaField("age", "INTEGER", mode="REQUIRED"),
    )

    records = [{"name": "Chip", "age": 2}, {"name": "Dale", "age": 3}]
    dataframe = pandas.DataFrame(records, columns=["name", "age"])
    table_id = "{}.{}.load_table_from_dataframe_w_required".format(
        bigquery_client.project, dataset_id
    )

    # Create the table before loading so that schema mismatch errors are
    # identified.
    table = helpers.retry_403(bigquery_client.create_table)(
        bigquery.Table(table_id, schema=table_schema)
    )

    load_job = bigquery_client.load_table_from_dataframe(dataframe, table_id)
    load_job.result()

    table = bigquery_client.get_table(table)
    assert tuple(table.schema) == table_schema
    assert table.num_rows == 2
    for field in table.schema:
        assert field.mode == "REQUIRED"


def test_load_table_from_dataframe_w_required_but_local_nulls_fails(
    bigquery_client, dataset_id
):
    """Test that a DataFrame with nulls can't be uploaded to a table with
    required columns.

    See: https://github.com/googleapis/python-bigquery/issues/1692
    """
    table_schema = (
        bigquery.SchemaField("name", "STRING", mode="REQUIRED"),
        bigquery.SchemaField("age", "INTEGER", mode="REQUIRED"),
    )

    records = [
        {"name": "Chip", "age": 2},
        {"name": "Dale", "age": 3},
        {"name": None, "age": None},
        {"name": "Alvin", "age": 4},
    ]
    dataframe = pandas.DataFrame(records, columns=["name", "age"])
    table_id = (
        "{}.{}.load_table_from_dataframe_w_required_but_local_nulls_fails".format(
            bigquery_client.project, dataset_id
        )
    )

    # Create the table before loading so that schema mismatch errors are
    # identified.
    helpers.retry_403(bigquery_client.create_table)(
        bigquery.Table(table_id, schema=table_schema)
    )

    with pytest.raises(google.api_core.exceptions.BadRequest, match="null"):
        bigquery_client.load_table_from_dataframe(dataframe, table_id).result()


def test_load_table_from_dataframe_w_explicit_schema(bigquery_client, dataset_id):
    # Schema with all scalar types.
    # See:
    #       https://github.com/googleapis/python-bigquery/issues/61
    #       https://issuetracker.google.com/issues/151765076
    table_schema = (
        bigquery.SchemaField("row_num", "INTEGER"),
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("bytes_col", "BYTES"),
        bigquery.SchemaField("date_col", "DATE"),
        bigquery.SchemaField("dt_col", "DATETIME"),
        bigquery.SchemaField("float_col", "FLOAT"),
        bigquery.SchemaField("geo_col", "GEOGRAPHY"),
        bigquery.SchemaField("int_col", "INTEGER"),
        bigquery.SchemaField("num_col", "NUMERIC"),
        bigquery.SchemaField("str_col", "STRING"),
        bigquery.SchemaField("time_col", "TIME"),
        bigquery.SchemaField("ts_col", "TIMESTAMP"),
        bigquery.SchemaField("bignum_col", "BIGNUMERIC"),
    )

    df_data = [
        ("row_num", [1, 2, 3]),
        ("bool_col", [True, None, False]),
        ("bytes_col", [b"abc", None, b"def"]),
        ("date_col", [datetime.date(1, 1, 1), None, datetime.date(9999, 12, 31)]),
        (
            "dt_col",
            [
                datetime.datetime(1, 1, 1, 0, 0, 0),
                None,
                datetime.datetime(9999, 12, 31, 23, 59, 59, 999999),
            ],
        ),
        ("float_col", [float("-inf"), float("nan"), float("inf")]),
        (
            "geo_col",
            ["POINT(30 10)", None, "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"],
        ),
        ("int_col", [-9223372036854775808, None, 9223372036854775807]),
        (
            "num_col",
            [
                decimal.Decimal("-99999999999999999999999999999.999999999"),
                None,
                decimal.Decimal("99999999999999999999999999999.999999999"),
            ],
        ),
        ("str_col", ["abc", None, "def"]),
        (
            "time_col",
            [datetime.time(0, 0, 0), None, datetime.time(23, 59, 59, 999999)],
        ),
        (
            "ts_col",
            [
                datetime.datetime(1, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                None,
                datetime.datetime(
                    9999, 12, 31, 23, 59, 59, 999999, tzinfo=datetime.timezone.utc
                ),
            ],
        ),
        (
            "bignum_col",
            [
                decimal.Decimal("-{d38}.{d38}".format(d38="9" * 38)),
                None,
                decimal.Decimal("{d38}.{d38}".format(d38="9" * 38)),
            ],
        ),
    ]
    df_data = collections.OrderedDict(df_data)
    dataframe = pandas.DataFrame(df_data, dtype="object", columns=df_data.keys())

    table_id = "{}.{}.load_table_from_dataframe_w_explicit_schema".format(
        bigquery_client.project, dataset_id
    )

    job_config = bigquery.LoadJobConfig(schema=table_schema)
    load_job = bigquery_client.load_table_from_dataframe(
        dataframe, table_id, job_config=job_config
    )
    load_job.result()

    table = bigquery_client.get_table(table_id)
    assert tuple(table.schema) == table_schema
    assert table.num_rows == 3

    result = bigquery_client.list_rows(table).to_dataframe()
    result.sort_values("row_num", inplace=True)

    # Check that extreme DATE/DATETIME values are loaded correctly.
    # https://github.com/googleapis/python-bigquery/issues/1076
    assert result["date_col"][0] == datetime.date(1, 1, 1)
    assert result["date_col"][2] == datetime.date(9999, 12, 31)
    assert result["dt_col"][0] == datetime.datetime(1, 1, 1, 0, 0, 0)
    assert result["dt_col"][2] == datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)
    assert result["ts_col"][0] == datetime.datetime(
        1, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
    )
    assert result["ts_col"][2] == datetime.datetime(
        9999, 12, 31, 23, 59, 59, 999999, tzinfo=datetime.timezone.utc
    )


def test_load_table_from_dataframe_w_struct_datatype(bigquery_client, dataset_id):
    """Test that a DataFrame with struct datatype can be uploaded if a
    BigQuery schema is specified.

    https://github.com/googleapis/python-bigquery/issues/21
    """
    table_id = "{}.{}.load_table_from_dataframe_w_struct_datatype".format(
        bigquery_client.project, dataset_id
    )
    table_schema = [
        bigquery.SchemaField(
            "bar",
            "RECORD",
            fields=[
                bigquery.SchemaField("id", "INTEGER", mode="REQUIRED"),
                bigquery.SchemaField("age", "INTEGER", mode="REQUIRED"),
            ],
            mode="REQUIRED",
        ),
    ]
    table = helpers.retry_403(bigquery_client.create_table)(
        bigquery.Table(table_id, schema=table_schema)
    )

    df_data = [{"id": 1, "age": 21}, {"id": 2, "age": 22}, {"id": 2, "age": 23}]
    dataframe = pandas.DataFrame(data={"bar": df_data}, columns=["bar"])

    load_job = bigquery_client.load_table_from_dataframe(dataframe, table_id)
    load_job.result()

    table = bigquery_client.get_table(table_id)
    assert table.schema == table_schema
    assert table.num_rows == 3


def test_load_table_from_dataframe_w_explicit_schema_source_format_csv(
    bigquery_client, dataset_id
):
    from arrivy.google.cloud.bigquery.job import SourceFormat

    table_schema = (
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("bytes_col", "BYTES"),
        bigquery.SchemaField("date_col", "DATE"),
        bigquery.SchemaField("dt_col", "DATETIME"),
        bigquery.SchemaField("float_col", "FLOAT"),
        bigquery.SchemaField("geo_col", "GEOGRAPHY"),
        bigquery.SchemaField("int_col", "INTEGER"),
        bigquery.SchemaField("num_col", "NUMERIC"),
        bigquery.SchemaField("bignum_col", "BIGNUMERIC"),
        bigquery.SchemaField("str_col", "STRING"),
        bigquery.SchemaField("time_col", "TIME"),
        bigquery.SchemaField("ts_col", "TIMESTAMP"),
    )
    df_data = collections.OrderedDict(
        [
            ("bool_col", [True, None, False]),
            ("bytes_col", ["abc", None, "def"]),
            (
                "date_col",
                [datetime.date(1, 1, 1), None, datetime.date(9999, 12, 31)],
            ),
            (
                "dt_col",
                [
                    datetime.datetime(1, 1, 1, 0, 0, 0),
                    None,
                    datetime.datetime(9999, 12, 31, 23, 59, 59, 999999),
                ],
            ),
            ("float_col", [float("-inf"), float("nan"), float("inf")]),
            (
                "geo_col",
                [
                    "POINT(30 10)",
                    None,
                    "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))",
                ],
            ),
            ("int_col", [-9223372036854775808, None, 9223372036854775807]),
            (
                "num_col",
                [
                    decimal.Decimal("-99999999999999999999999999999.999999999"),
                    None,
                    decimal.Decimal("99999999999999999999999999999.999999999"),
                ],
            ),
            (
                "bignum_col",
                [
                    decimal.Decimal("-{d38}.{d38}".format(d38="9" * 38)),
                    None,
                    decimal.Decimal("{d38}.{d38}".format(d38="9" * 38)),
                ],
            ),
            ("str_col", ["abc", None, "def"]),
            (
                "time_col",
                [datetime.time(0, 0, 0), None, datetime.time(23, 59, 59, 999999)],
            ),
            (
                "ts_col",
                [
                    datetime.datetime(1, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                    None,
                    datetime.datetime(
                        9999, 12, 31, 23, 59, 59, 999999, tzinfo=datetime.timezone.utc
                    ),
                ],
            ),
        ]
    )
    dataframe = pandas.DataFrame(df_data, dtype="object", columns=df_data.keys())

    table_id = "{}.{}.load_table_from_dataframe_w_explicit_schema_csv".format(
        bigquery_client.project, dataset_id
    )

    job_config = bigquery.LoadJobConfig(
        schema=table_schema, source_format=SourceFormat.CSV
    )
    load_job = bigquery_client.load_table_from_dataframe(
        dataframe, table_id, job_config=job_config
    )
    load_job.result()

    table = bigquery_client.get_table(table_id)
    assert tuple(table.schema) == table_schema
    assert table.num_rows == 3


def test_load_table_from_dataframe_w_explicit_schema_source_format_csv_floats(
    bigquery_client, dataset_id, table_id
):
    from arrivy.google.cloud.bigquery.job import SourceFormat

    table_schema = (bigquery.SchemaField("float_col", "FLOAT"),)
    df_data = collections.OrderedDict(
        [
            (
                "float_col",
                [
                    0.14285714285714285,
                    0.51428571485748,
                    0.87128748,
                    1.807960649,
                    2.0679610649,
                    2.4406779661016949,
                    3.7148514257,
                    3.8571428571428572,
                    1.51251252e40,
                ],
            ),
        ]
    )
    dataframe = pandas.DataFrame(df_data, dtype="object", columns=df_data.keys())

    job_config = bigquery.LoadJobConfig(
        schema=table_schema, source_format=SourceFormat.CSV
    )
    load_job = bigquery_client.load_table_from_dataframe(
        dataframe, table_id, job_config=job_config
    )
    load_job.result()

    table = bigquery_client.get_table(table_id)
    rows = bigquery_client.list_rows(table_id)
    floats = [r.values()[0] for r in rows]
    assert tuple(table.schema) == table_schema
    assert table.num_rows == 9
    assert floats == df_data["float_col"]


def test_query_results_to_dataframe(bigquery_client):
    QUERY = """
    SELECT id, `by`, timestamp, dead
    FROM `bigquery-public-data.hacker_news.full`
    LIMIT 10
    """

    df = bigquery_client.query(QUERY).result().to_dataframe()

    assert isinstance(df, pandas.DataFrame)
    assert len(df) == 10  # verify the number of rows
    column_names = ["id", "by", "timestamp", "dead"]
    assert list(df) == column_names  # verify the column names
    exp_datatypes = {
        "id": int,
        "by": str,
        "timestamp": pandas.Timestamp,
        "dead": bool,
    }
    for _, row in df.iterrows():
        for col in column_names:
            # all the schema fields are nullable, so None is acceptable
            if not pandas.isna(row[col]):
                assert isinstance(row[col], exp_datatypes[col])


def test_query_results_to_dataframe_w_bqstorage(bigquery_client):
    query = """
    SELECT id, `by`, timestamp, dead
    FROM `bigquery-public-data.hacker_news.full`
    LIMIT 10
    """

    bqstorage_client = bigquery_storage.BigQueryReadClient(
        credentials=bigquery_client._credentials
    )

    df = bigquery_client.query(query).result().to_dataframe(bqstorage_client)

    assert isinstance(df, pandas.DataFrame)
    assert len(df) == 10  # verify the number of rows
    column_names = ["id", "by", "timestamp", "dead"]
    assert list(df) == column_names
    exp_datatypes = {
        "id": int,
        "by": str,
        "timestamp": pandas.Timestamp,
        "dead": bool,
    }
    for index, row in df.iterrows():
        for col in column_names:
            # all the schema fields are nullable, so None is acceptable
            if not pandas.isna(row[col]):
                assert isinstance(row[col], exp_datatypes[col])


def test_insert_rows_from_dataframe(bigquery_client, dataset_id):
    SF = bigquery.SchemaField
    schema = [
        SF("float_col", "FLOAT", mode="REQUIRED"),
        SF("int_col", "INTEGER", mode="REQUIRED"),
        SF("bool_col", "BOOLEAN", mode="REQUIRED"),
        SF("string_col", "STRING", mode="NULLABLE"),
        SF("date_col", "DATE", mode="NULLABLE"),
        SF("time_col", "TIME", mode="NULLABLE"),
    ]

    dataframe = pandas.DataFrame(
        [
            {
                "float_col": 1.11,
                "bool_col": True,
                "string_col": "my string",
                "int_col": 10,
                "date_col": datetime.date(2021, 1, 1),
                "time_col": datetime.time(21, 1, 1),
            },
            {
                "float_col": 2.22,
                "bool_col": False,
                "string_col": "another string",
                "int_col": 20,
                "date_col": datetime.date(2021, 1, 2),
                "time_col": datetime.time(21, 1, 2),
            },
            {
                "float_col": 3.33,
                "bool_col": False,
                "string_col": "another string",
                "int_col": 30,
                "date_col": datetime.date(2021, 1, 3),
                "time_col": datetime.time(21, 1, 3),
            },
            {
                "float_col": 4.44,
                "bool_col": True,
                "string_col": "another string",
                "int_col": 40,
                "date_col": datetime.date(2021, 1, 4),
                "time_col": datetime.time(21, 1, 4),
            },
            {
                "float_col": 5.55,
                "bool_col": False,
                "string_col": "another string",
                "int_col": 50,
                "date_col": datetime.date(2021, 1, 5),
                "time_col": datetime.time(21, 1, 5),
            },
            {
                "float_col": 6.66,
                "bool_col": True,
                # Include a NaN value, because pandas often uses NaN as a
                # NULL value indicator.
                "string_col": float("NaN"),
                "int_col": 60,
                "date_col": datetime.date(2021, 1, 6),
                "time_col": datetime.time(21, 1, 6),
            },
        ]
    )
    dataframe["date_col"] = dataframe["date_col"].astype("dbdate")
    dataframe["time_col"] = dataframe["time_col"].astype("dbtime")

    table_id = f"{bigquery_client.project}.{dataset_id}.test_insert_rows_from_dataframe"
    table_arg = bigquery.Table(table_id, schema=schema)
    table = helpers.retry_403(bigquery_client.create_table)(table_arg)

    chunk_errors = bigquery_client.insert_rows_from_dataframe(
        table, dataframe, chunk_size=3
    )
    for errors in chunk_errors:
        assert not errors
    expected = [
        # Pandas often represents NULL values as NaN. Convert to None for
        # easier comparison.
        tuple(None if col != col else col for col in data_row)
        for data_row in dataframe.itertuples(index=False)
    ]

    # Use query to fetch rows instead of listing directly from the table so
    # that we get values from the streaming buffer "within a few seconds".
    # https://cloud.google.com/bigquery/streaming-data-into-bigquery#dataavailability
    @google.api_core.retry.Retry(
        predicate=google.api_core.retry.if_exception_type(MissingDataError)
    )
    def get_rows():
        rows = list(
            bigquery_client.query(
                "SELECT * FROM `{}.{}.{}`".format(
                    table.project, table.dataset_id, table.table_id
                )
            )
        )
        if len(rows) != len(expected):
            raise MissingDataError()
        return rows

    rows = get_rows()
    sorted_rows = sorted(rows, key=operator.attrgetter("int_col"))
    row_tuples = [r.values() for r in sorted_rows]

    for row, expected_row in zip(row_tuples, expected):
        assert (
            # Use Counter to verify the same number of values in each, because
            # column order does not matter.
            collections.Counter(row)
            == collections.Counter(expected_row)
        )


def test_nested_table_to_dataframe(bigquery_client, dataset_id):
    from arrivy.google.cloud.bigquery.job import SourceFormat
    from arrivy.google.cloud.bigquery.job import WriteDisposition

    SF = bigquery.SchemaField
    schema = [
        SF("string_col", "STRING", mode="NULLABLE"),
        SF(
            "record_col",
            "RECORD",
            mode="NULLABLE",
            fields=[
                SF("nested_string", "STRING", mode="NULLABLE"),
                SF("nested_repeated", "INTEGER", mode="REPEATED"),
                SF(
                    "nested_record",
                    "RECORD",
                    mode="NULLABLE",
                    fields=[SF("nested_nested_string", "STRING", mode="NULLABLE")],
                ),
            ],
        ),
        SF("bigfloat_col", "FLOAT", mode="NULLABLE"),
        SF("smallfloat_col", "FLOAT", mode="NULLABLE"),
    ]
    record = {
        "nested_string": "another string value",
        "nested_repeated": [0, 1, 2],
        "nested_record": {"nested_nested_string": "some deep insight"},
    }
    to_insert = [
        {
            "string_col": "Some value",
            "record_col": record,
            "bigfloat_col": 3.14,
            "smallfloat_col": 2.72,
        }
    ]
    rows = [json.dumps(row) for row in to_insert]
    body = io.BytesIO("{}\n".format("\n".join(rows)).encode("ascii"))
    table_id = f"{bigquery_client.project}.{dataset_id}.test_nested_table_to_dataframe"
    job_config = bigquery.LoadJobConfig()
    job_config.write_disposition = WriteDisposition.WRITE_TRUNCATE
    job_config.source_format = SourceFormat.NEWLINE_DELIMITED_JSON
    job_config.schema = schema
    # Load a table using a local JSON file from memory.
    bigquery_client.load_table_from_file(body, table_id, job_config=job_config).result()

    df = bigquery_client.list_rows(table_id, selected_fields=schema).to_dataframe(
        dtypes={"smallfloat_col": "float16"}
    )

    assert isinstance(df, pandas.DataFrame)
    assert len(df) == 1  # verify the number of rows
    exp_columns = ["string_col", "record_col", "bigfloat_col", "smallfloat_col"]
    assert list(df) == exp_columns  # verify the column names
    row = df.iloc[0]
    # verify the row content
    assert row["string_col"] == "Some value"
    expected_keys = tuple(sorted(record.keys()))
    row_keys = tuple(sorted(row["record_col"].keys()))
    assert row_keys == expected_keys
    # Can't compare numpy arrays, which pyarrow encodes the embedded
    # repeated column to, so convert to list.
    assert list(row["record_col"]["nested_repeated"]) == [0, 1, 2]
    # verify that nested data can be accessed with indices/keys
    assert row["record_col"]["nested_repeated"][0] == 0
    assert (
        row["record_col"]["nested_record"]["nested_nested_string"]
        == "some deep insight"
    )
    # verify dtypes
    assert df.dtypes["bigfloat_col"].name == "float64"
    assert df.dtypes["smallfloat_col"].name == "float16"


def test_list_rows_max_results_w_bqstorage(bigquery_client):
    table_ref = bigquery.DatasetReference("bigquery-public-data", "utility_us").table(
        "country_code_iso"
    )
    bqstorage_client = bigquery_storage.BigQueryReadClient(
        credentials=bigquery_client._credentials
    )

    row_iterator = bigquery_client.list_rows(
        table_ref,
        selected_fields=[bigquery.SchemaField("country_name", "STRING")],
        max_results=100,
    )
    with pytest.warns(
        UserWarning, match="Cannot use bqstorage_client if max_results is set"
    ):
        dataframe = row_iterator.to_dataframe(bqstorage_client=bqstorage_client)

    assert len(dataframe.index) == 100


@pytest.mark.skipif(PANDAS_INSTALLED_VERSION[0:2] not in ["0.", "1."], reason="")
@pytest.mark.parametrize(
    ("max_results",),
    (
        (None,),
        (10,),
    ),  # Use BQ Storage API.  # Use REST API.
)
def test_list_rows_nullable_scalars_dtypes(bigquery_client, scalars_table, max_results):
    # TODO(GH#836): Avoid INTERVAL columns until they are supported by the
    # BigQuery Storage API and pyarrow.
    schema = [
        bigquery.SchemaField("bool_col", enums.SqlTypeNames.BOOLEAN),
        bigquery.SchemaField("bignumeric_col", enums.SqlTypeNames.BIGNUMERIC),
        bigquery.SchemaField("bytes_col", enums.SqlTypeNames.BYTES),
        bigquery.SchemaField("date_col", enums.SqlTypeNames.DATE),
        bigquery.SchemaField("datetime_col", enums.SqlTypeNames.DATETIME),
        bigquery.SchemaField("float64_col", enums.SqlTypeNames.FLOAT64),
        bigquery.SchemaField("geography_col", enums.SqlTypeNames.GEOGRAPHY),
        bigquery.SchemaField("int64_col", enums.SqlTypeNames.INT64),
        bigquery.SchemaField("numeric_col", enums.SqlTypeNames.NUMERIC),
        bigquery.SchemaField("string_col", enums.SqlTypeNames.STRING),
        bigquery.SchemaField("time_col", enums.SqlTypeNames.TIME),
        bigquery.SchemaField("timestamp_col", enums.SqlTypeNames.TIMESTAMP),
    ]

    df = bigquery_client.list_rows(
        scalars_table,
        max_results=max_results,
        selected_fields=schema,
    ).to_dataframe()

    assert df.dtypes["bool_col"].name == "boolean"
    assert df.dtypes["datetime_col"].name == "datetime64[ns]"
    assert df.dtypes["float64_col"].name == "float64"
    assert df.dtypes["int64_col"].name == "Int64"
    assert df.dtypes["timestamp_col"].name == "datetime64[ns, UTC]"
    assert df.dtypes["date_col"].name == "dbdate"
    assert df.dtypes["time_col"].name == "dbtime"

    # decimal.Decimal is used to avoid loss of precision.
    assert df.dtypes["bignumeric_col"].name == "object"
    assert df.dtypes["numeric_col"].name == "object"

    # pandas uses Python string and bytes objects.
    assert df.dtypes["bytes_col"].name == "object"
    assert df.dtypes["string_col"].name == "object"


@pytest.mark.parametrize(
    ("max_results",),
    (
        (None,),
        (10,),
    ),  # Use BQ Storage API.  # Use REST API.
)
def test_list_rows_nullable_scalars_extreme_dtypes(
    bigquery_client, scalars_extreme_table, max_results
):
    # TODO(GH#836): Avoid INTERVAL columns until they are supported by the
    # BigQuery Storage API and pyarrow.
    schema = [
        bigquery.SchemaField("bool_col", enums.SqlTypeNames.BOOLEAN),
        bigquery.SchemaField("bignumeric_col", enums.SqlTypeNames.BIGNUMERIC),
        bigquery.SchemaField("bytes_col", enums.SqlTypeNames.BYTES),
        bigquery.SchemaField("date_col", enums.SqlTypeNames.DATE),
        bigquery.SchemaField("datetime_col", enums.SqlTypeNames.DATETIME),
        bigquery.SchemaField("float64_col", enums.SqlTypeNames.FLOAT64),
        bigquery.SchemaField("geography_col", enums.SqlTypeNames.GEOGRAPHY),
        bigquery.SchemaField("int64_col", enums.SqlTypeNames.INT64),
        bigquery.SchemaField("numeric_col", enums.SqlTypeNames.NUMERIC),
        bigquery.SchemaField("string_col", enums.SqlTypeNames.STRING),
        bigquery.SchemaField("time_col", enums.SqlTypeNames.TIME),
        bigquery.SchemaField("timestamp_col", enums.SqlTypeNames.TIMESTAMP),
    ]

    df = bigquery_client.list_rows(
        scalars_extreme_table,
        max_results=max_results,
        selected_fields=schema,
    ).to_dataframe()

    # Extreme values are out-of-bounds for pandas datetime64 values, which use
    # nanosecond precision.  Values before 1677-09-21 and after 2262-04-11 must
    # be represented with object.
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timestamp-limitations
    assert df.dtypes["date_col"].name == "object"
    assert df.dtypes["datetime_col"].name == "object"
    assert df.dtypes["timestamp_col"].name == "object"

    # These pandas dtypes can handle the same ranges as BigQuery.
    assert df.dtypes["bool_col"].name == "boolean"
    assert df.dtypes["float64_col"].name == "float64"
    assert df.dtypes["int64_col"].name == "Int64"
    assert df.dtypes["time_col"].name == "dbtime"

    # decimal.Decimal is used to avoid loss of precision.
    assert df.dtypes["numeric_col"].name == "object"
    assert df.dtypes["bignumeric_col"].name == "object"

    # pandas uses Python string and bytes objects.
    assert df.dtypes["bytes_col"].name == "object"
    assert df.dtypes["string_col"].name == "object"


@pytest.mark.parametrize(
    ("max_results",),
    (
        (None,),
        (10,),
    ),  # Use BQ Storage API.  # Use REST API.
)
def test_list_rows_nullable_scalars_extreme_dtypes_w_custom_dtype(
    bigquery_client, scalars_extreme_table, max_results
):
    # TODO(GH#836): Avoid INTERVAL columns until they are supported by the
    # BigQuery Storage API and pyarrow.
    schema = [
        bigquery.SchemaField("bool_col", enums.SqlTypeNames.BOOLEAN),
        bigquery.SchemaField("bignumeric_col", enums.SqlTypeNames.BIGNUMERIC),
        bigquery.SchemaField("bytes_col", enums.SqlTypeNames.BYTES),
        bigquery.SchemaField("date_col", enums.SqlTypeNames.DATE),
        bigquery.SchemaField("datetime_col", enums.SqlTypeNames.DATETIME),
        bigquery.SchemaField("float64_col", enums.SqlTypeNames.FLOAT64),
        bigquery.SchemaField("geography_col", enums.SqlTypeNames.GEOGRAPHY),
        bigquery.SchemaField("int64_col", enums.SqlTypeNames.INT64),
        bigquery.SchemaField("numeric_col", enums.SqlTypeNames.NUMERIC),
        bigquery.SchemaField("string_col", enums.SqlTypeNames.STRING),
        bigquery.SchemaField("time_col", enums.SqlTypeNames.TIME),
        bigquery.SchemaField("timestamp_col", enums.SqlTypeNames.TIMESTAMP),
    ]

    df = bigquery_client.list_rows(
        scalars_extreme_table,
        max_results=max_results,
        selected_fields=schema,
    ).to_dataframe(
        bool_dtype=pandas.BooleanDtype(),
        int_dtype=pandas.Int64Dtype(),
        float_dtype=(
            pandas.Float64Dtype()
            if hasattr(pandas, "Float64Dtype")
            else pandas.StringDtype()
        ),
        string_dtype=pandas.StringDtype(),
        date_dtype=(
            pandas.ArrowDtype(pyarrow.date32())
            if hasattr(pandas, "ArrowDtype")
            else None
        ),
        datetime_dtype=(
            pandas.ArrowDtype(pyarrow.timestamp("us"))
            if hasattr(pandas, "ArrowDtype")
            else None
        ),
        time_dtype=(
            pandas.ArrowDtype(pyarrow.time64("us"))
            if hasattr(pandas, "ArrowDtype")
            else None
        ),
        timestamp_dtype=(
            pandas.ArrowDtype(pyarrow.timestamp("us", tz="UTC"))
            if hasattr(pandas, "ArrowDtype")
            else None
        ),
    )

    # These pandas dtypes are handled by the custom dtypes.
    assert df.dtypes["bool_col"].name == "boolean"
    assert df.dtypes["float64_col"].name == "Float64"
    assert df.dtypes["int64_col"].name == "Int64"
    assert df.dtypes["string_col"].name == "string"

    assert (
        df.dtypes["date_col"].name == "date32[day][pyarrow]"
        if hasattr(pandas, "ArrowDtype")
        else "datetime64[ns]"
    )
    assert (
        df.dtypes["datetime_col"].name == "timestamp[us][pyarrow]"
        if hasattr(pandas, "ArrowDtype")
        else "object"
    )
    assert (
        df.dtypes["timestamp_col"].name == "timestamp[us, tz=UTC][pyarrow]"
        if hasattr(pandas, "ArrowDtype")
        else "object"
    )
    assert (
        df.dtypes["time_col"].name == "time64[us][pyarrow]"
        if hasattr(pandas, "ArrowDtype")
        else "object"
    )

    # decimal.Decimal is used to avoid loss of precision.
    assert df.dtypes["numeric_col"].name == "object"
    assert df.dtypes["bignumeric_col"].name == "object"

    # pandas uses Python bytes objects.
    assert df.dtypes["bytes_col"].name == "object"


def test_upload_time_and_datetime_56(bigquery_client, dataset_id):
    df = pandas.DataFrame(
        dict(
            dt=[
                datetime.datetime(2020, 1, 8, 8, 0, 0),
                datetime.datetime(
                    2020,
                    1,
                    8,
                    8,
                    0,
                    0,
                    tzinfo=datetime.timezone(datetime.timedelta(hours=-7)),
                ),
            ],
            t=[datetime.time(0, 0, 10, 100001), None],
        )
    )
    table = f"{dataset_id}.test_upload_time_and_datetime"
    bigquery_client.load_table_from_dataframe(df, table).result()
    data = list(map(list, bigquery_client.list_rows(table)))
    assert data == [
        [
            datetime.datetime(2020, 1, 8, 8, 0, tzinfo=datetime.timezone.utc),
            datetime.time(0, 0, 10, 100001),
        ],
        [datetime.datetime(2020, 1, 8, 15, 0, tzinfo=datetime.timezone.utc), None],
    ]

    from arrivy.google.cloud.bigquery import job, schema

    table = f"{dataset_id}.test_upload_time_and_datetime_dt"
    config = job.LoadJobConfig(
        schema=[schema.SchemaField("dt", "DATETIME"), schema.SchemaField("t", "TIME")]
    )

    bigquery_client.load_table_from_dataframe(df, table, job_config=config).result()
    data = list(map(list, bigquery_client.list_rows(table)))
    assert data == [
        [datetime.datetime(2020, 1, 8, 8, 0), datetime.time(0, 0, 10, 100001)],
        [datetime.datetime(2020, 1, 8, 15, 0), None],
    ]


def test_to_dataframe_geography_as_objects(bigquery_client, dataset_id):
    wkt = pytest.importorskip("shapely.wkt")
    bigquery_client.query(
        f"create table {dataset_id}.lake (name string, geog geography)"
    ).result()
    bigquery_client.query(
        f"""
        insert into {dataset_id}.lake (name, geog) values
        ('foo', st_geogfromtext('point(0 0)')),
        ('bar', st_geogfromtext('point(0 1)')),
        ('baz', null)
        """
    ).result()
    df = bigquery_client.query(
        f"select * from {dataset_id}.lake order by name"
    ).to_dataframe(geography_as_object=True)
    assert list(df["name"]) == ["bar", "baz", "foo"]
    assert df["geog"][0] == wkt.loads("point(0 1)")
    assert pandas.isna(df["geog"][1])
    assert df["geog"][2] == wkt.loads("point(0 0)")


def test_to_geodataframe(bigquery_client, dataset_id):
    geopandas = pytest.importorskip("geopandas")
    from shapely import wkt

    bigquery_client.query(
        f"create table {dataset_id}.geolake (name string, geog geography)"
    ).result()
    bigquery_client.query(
        f"""
        insert into {dataset_id}.geolake (name, geog) values
        ('foo', st_geogfromtext('point(0 0)')),
        ('bar', st_geogfromtext('polygon((0 0, 1 0, 1 1, 0 0))')),
        ('baz', null)
        """
    ).result()
    df = bigquery_client.query(
        f"select * from {dataset_id}.geolake order by name"
    ).to_geodataframe()
    assert df["geog"][0] == wkt.loads("polygon((0 0, 1 0, 1 1, 0 0))")
    assert pandas.isna(df["geog"][1])
    assert df["geog"][2] == wkt.loads("point(0 0)")
    assert isinstance(df, geopandas.GeoDataFrame)
    assert isinstance(df["geog"], geopandas.GeoSeries)

    with warnings.catch_warnings():
        # Computing the area on a GeoDataFrame that uses a geographic Coordinate
        # Reference System (CRS) produces a warning that we are not interested in.
        # We do not mind if the computed area is incorrect with respect to the
        # GeoDataFrame data, as long as it matches the expected "incorrect" value.
        warnings.filterwarnings("ignore", category=UserWarning)
        assert df.area[0] == 0.5
        assert pandas.isna(df.area[1])
        assert df.area[2] == 0.0

    assert df.crs.srs == "EPSG:4326"
    assert df.crs.name == "WGS 84"
    assert df.geog.crs.srs == "EPSG:4326"
    assert df.geog.crs.name == "WGS 84"


def test_load_geodataframe(bigquery_client, dataset_id):
    geopandas = pytest.importorskip("geopandas")
    import pandas
    from shapely import wkt
    from arrivy.google.cloud.bigquery.schema import SchemaField

    df = geopandas.GeoDataFrame(
        pandas.DataFrame(
            dict(
                name=["foo", "bar"],
                geo1=[None, None],
                geo2=[None, wkt.loads("Point(1 1)")],
            )
        ),
        geometry="geo1",
    )

    table_id = f"{dataset_id}.lake_from_gp"
    bigquery_client.load_table_from_dataframe(df, table_id).result()

    table = bigquery_client.get_table(table_id)
    assert table.schema == [
        SchemaField("name", "STRING", "NULLABLE"),
        SchemaField("geo1", "GEOGRAPHY", "NULLABLE"),
        SchemaField("geo2", "GEOGRAPHY", "NULLABLE"),
    ]
    assert sorted(map(list, bigquery_client.list_rows(table_id))) == [
        ["bar", None, "POINT(1 1)"],
        ["foo", None, None],
    ]


def test_load_dataframe_w_shapely(bigquery_client, dataset_id):
    wkt = pytest.importorskip("shapely.wkt")
    from arrivy.google.cloud.bigquery.schema import SchemaField

    df = pandas.DataFrame(
        dict(name=["foo", "bar"], geo=[None, wkt.loads("Point(1 1)")])
    )

    table_id = f"{dataset_id}.lake_from_shapes"
    bigquery_client.load_table_from_dataframe(df, table_id).result()

    table = bigquery_client.get_table(table_id)
    assert table.schema == [
        SchemaField("name", "STRING", "NULLABLE"),
        SchemaField("geo", "GEOGRAPHY", "NULLABLE"),
    ]
    assert sorted(map(list, bigquery_client.list_rows(table_id))) == [
        ["bar", "POINT(1 1)"],
        ["foo", None],
    ]

    bigquery_client.load_table_from_dataframe(df, table_id).result()
    assert sorted(map(list, bigquery_client.list_rows(table_id))) == [
        ["bar", "POINT(1 1)"],
        ["bar", "POINT(1 1)"],
        ["foo", None],
        ["foo", None],
    ]


def test_load_dataframe_w_wkb(bigquery_client, dataset_id):
    wkt = pytest.importorskip("shapely.wkt")
    from shapely import wkb
    from arrivy.google.cloud.bigquery.schema import SchemaField

    df = pandas.DataFrame(
        dict(name=["foo", "bar"], geo=[None, wkb.dumps(wkt.loads("Point(1 1)"))])
    )

    table_id = f"{dataset_id}.lake_from_wkb"
    # We create the table first, to inform the interpretation of the wkb data
    bigquery_client.query(
        f"create table {table_id} (name string, geo GEOGRAPHY)"
    ).result()
    bigquery_client.load_table_from_dataframe(df, table_id).result()

    table = bigquery_client.get_table(table_id)
    assert table.schema == [
        SchemaField("name", "STRING", "NULLABLE"),
        SchemaField("geo", "GEOGRAPHY", "NULLABLE"),
    ]
    assert sorted(map(list, bigquery_client.list_rows(table_id))) == [
        ["bar", "POINT(1 1)"],
        ["foo", None],
    ]



# Copyright 2015 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import mock


class TestEncryptionConfiguration(unittest.TestCase):
    KMS_KEY_NAME = "projects/1/locations/us/keyRings/1/cryptoKeys/1"

    @staticmethod
    def _get_target_class():
        from arrivy.google.cloud.bigquery.encryption_configuration import (
            EncryptionConfiguration,
        )

        return EncryptionConfiguration

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_ctor_defaults(self):
        encryption_config = self._make_one()
        self.assertIsNone(encryption_config.kms_key_name)

    def test_ctor_with_key(self):
        encryption_config = self._make_one(kms_key_name=self.KMS_KEY_NAME)
        self.assertEqual(encryption_config.kms_key_name, self.KMS_KEY_NAME)

    def test_kms_key_name_setter(self):
        encryption_config = self._make_one()
        self.assertIsNone(encryption_config.kms_key_name)
        encryption_config.kms_key_name = self.KMS_KEY_NAME
        self.assertEqual(encryption_config.kms_key_name, self.KMS_KEY_NAME)
        encryption_config.kms_key_name = None
        self.assertIsNone(encryption_config.kms_key_name)

    def test_from_api_repr(self):
        RESOURCE = {"kmsKeyName": self.KMS_KEY_NAME}
        klass = self._get_target_class()
        encryption_config = klass.from_api_repr(RESOURCE)
        self.assertEqual(encryption_config.kms_key_name, self.KMS_KEY_NAME)

    def test_to_api_repr(self):
        encryption_config = self._make_one(kms_key_name=self.KMS_KEY_NAME)
        resource = encryption_config.to_api_repr()
        self.assertEqual(resource, {"kmsKeyName": self.KMS_KEY_NAME})

    def test___eq___wrong_type(self):
        encryption_config = self._make_one()
        other = object()
        self.assertNotEqual(encryption_config, other)
        self.assertEqual(encryption_config, mock.ANY)

    def test___eq___kms_key_name_mismatch(self):
        encryption_config = self._make_one()
        other = self._make_one(self.KMS_KEY_NAME)
        self.assertNotEqual(encryption_config, other)

    def test___eq___hit(self):
        encryption_config = self._make_one(self.KMS_KEY_NAME)
        other = self._make_one(self.KMS_KEY_NAME)
        self.assertEqual(encryption_config, other)

    def test___ne___wrong_type(self):
        encryption_config = self._make_one()
        other = object()
        self.assertNotEqual(encryption_config, other)
        self.assertEqual(encryption_config, mock.ANY)

    def test___ne___same_value(self):
        encryption_config1 = self._make_one(self.KMS_KEY_NAME)
        encryption_config2 = self._make_one(self.KMS_KEY_NAME)
        # unittest ``assertEqual`` uses ``==`` not ``!=``.
        comparison_val = encryption_config1 != encryption_config2
        self.assertFalse(comparison_val)

    def test___ne___different_values(self):
        encryption_config1 = self._make_one()
        encryption_config2 = self._make_one(self.KMS_KEY_NAME)
        self.assertNotEqual(encryption_config1, encryption_config2)

    def test___hash__set_equality(self):
        encryption_config1 = self._make_one(self.KMS_KEY_NAME)
        encryption_config2 = self._make_one(self.KMS_KEY_NAME)
        set_one = {encryption_config1, encryption_config2}
        set_two = {encryption_config1, encryption_config2}
        self.assertEqual(set_one, set_two)

    def test___hash__not_equals(self):
        encryption_config1 = self._make_one()
        encryption_config2 = self._make_one(self.KMS_KEY_NAME)
        set_one = {encryption_config1}
        set_two = {encryption_config2}
        self.assertNotEqual(set_one, set_two)

    def test___repr__(self):
        encryption_config = self._make_one(self.KMS_KEY_NAME)
        expected = "EncryptionConfiguration({})".format(self.KMS_KEY_NAME)
        self.assertEqual(repr(encryption_config), expected)
